from __future__ import annotations

from pathlib import Path

from gabion import config


def test_load_toml_missing_and_invalid(tmp_path: Path) -> None:
    missing = tmp_path / "missing.toml"
    assert config._load_toml(missing) == {}

    invalid = tmp_path / "invalid.toml"
    invalid.write_text("not = [toml", encoding="utf-8")
    assert config._load_toml(invalid) == {}

    assert config._load_toml(tmp_path) == {}


def test_load_config_default_path(tmp_path: Path) -> None:
    cfg = tmp_path / config.DEFAULT_CONFIG_NAME
    cfg.write_text("[dataflow]\nstrictness = 'low'\n", encoding="utf-8")
    data = config.load_config(root=tmp_path, config_path=None)
    assert data["dataflow"]["strictness"] == "low"



from __future__ import annotations

from pathlib import Path
import sys


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis import dataflow_audit as da

    return da


def _fn(da, *, name: str, qual: str, path: Path, class_name: str | None = None):
    return da.FunctionInfo(
        name=name,
        qual=qual,
        path=path,
        params=[],
        annots={},
        calls=[],
        unused_params=set(),
        class_name=class_name,
        scope=(),
        lexical_scope=(),
    )


def test_resolve_callee_globals_only() -> None:
    da = _load()
    path = Path("pkg/mod.py")
    caller = _fn(da, name="caller", qual="pkg.mod.caller", path=path)
    callee = _fn(da, name="target", qual="pkg.mod.target", path=path)
    by_name = {"target": [callee]}
    by_qual = {callee.qual: callee}
    resolved = da._resolve_callee("target", caller, by_name, by_qual)
    assert resolved is callee


def test_resolve_callee_ambiguous_then_global() -> None:
    da = _load()
    path = Path("pkg/mod.py")
    caller = da.FunctionInfo(
        name="caller",
        qual="pkg.mod.caller",
        path=path,
        params=[],
        annots={},
        calls=[],
        unused_params=set(),
        class_name=None,
        scope=(),
        lexical_scope=(),
    )
    scoped_one = da.FunctionInfo(
        name="target",
        qual="pkg.mod.outer.caller.target",
        path=path,
        params=[],
        annots={},
        calls=[],
        unused_params=set(),
        class_name=None,
        scope=("outer", "caller"),
        lexical_scope=("caller",),
    )
    scoped_two = da.FunctionInfo(
        name="target",
        qual="pkg.mod.outer.caller.target2",
        path=path,
        params=[],
        annots={},
        calls=[],
        unused_params=set(),
        class_name=None,
        scope=("outer", "caller"),
        lexical_scope=("caller",),
    )
    global_candidate = da.FunctionInfo(
        name="target",
        qual="pkg.mod.target",
        path=path,
        params=[],
        annots={},
        calls=[],
        unused_params=set(),
        class_name=None,
        scope=(),
        lexical_scope=(),
    )
    by_name = {"target": [scoped_one, scoped_two, global_candidate]}
    by_qual = {
        scoped_one.qual: scoped_one,
        scoped_two.qual: scoped_two,
        global_candidate.qual: global_candidate,
    }
    resolved = da._resolve_callee("target", caller, by_name, by_qual)
    assert resolved is global_candidate


def test_resolve_callee_import_and_star() -> None:
    da = _load()
    path = Path("pkg/mod.py")
    caller = _fn(da, name="caller", qual="pkg.mod.caller", path=path)
    callee = _fn(da, name="Imported", qual="pkg.other.fn", path=Path("pkg/other.py"))
    by_name = {"Imported": [callee]}
    by_qual = {callee.qual: callee}
    table = da.SymbolTable()
    table.internal_roots.add("pkg")
    table.imports[("pkg.mod", "Imported")] = "pkg.other.fn"
    resolved = da._resolve_callee(
        "Imported",
        caller,
        by_name,
        by_qual,
        symbol_table=table,
        project_root=None,
    )
    assert resolved is callee

    table.star_imports["pkg.mod"] = {"pkg.star"}
    table.module_exports["pkg.star"] = {"StarFunc"}
    by_qual["pkg.star.StarFunc"] = _fn(
        da,
        name="StarFunc",
        qual="pkg.star.StarFunc",
        path=Path("pkg/star.py"),
    )
    resolved_star = da._resolve_callee(
        "StarFunc",
        caller,
        by_name,
        by_qual,
        symbol_table=table,
        project_root=None,
    )
    assert resolved_star is by_qual["pkg.star.StarFunc"]


def test_resolve_callee_import_filtered_out() -> None:
    da = _load()
    path = Path("pkg/mod.py")
    caller = _fn(da, name="caller", qual="pkg.mod.caller", path=path)
    callee = _fn(
        da,
        name="Ext",
        qual="ext.lib.fn",
        path=Path("ext/lib.py"),
        class_name="ExtClass",
    )
    by_name = {"Ext": [callee]}
    by_qual = {callee.qual: callee}
    table = da.SymbolTable(external_filter=True)
    table.imports[("pkg.mod", "Ext")] = "ext.lib.fn"
    resolved = da._resolve_callee(
        "Ext",
        caller,
        by_name,
        by_qual,
        symbol_table=table,
        project_root=None,
    )
    assert resolved is None


def test_resolve_callee_self_and_base_import() -> None:
    da = _load()
    path = Path("pkg/mod.py")
    caller = _fn(da, name="caller", qual="pkg.mod.Service.caller", path=path, class_name="Service")
    self_method = _fn(da, name="run", qual="pkg.mod.Service.run", path=path, class_name="Service")
    by_name = {"run": [self_method]}
    by_qual = {self_method.qual: self_method}
    resolved = da._resolve_callee(
        "self.run",
        caller,
        by_name,
        by_qual,
        symbol_table=da.SymbolTable(),
        project_root=None,
    )
    assert resolved is self_method

    table = da.SymbolTable()
    table.internal_roots.add("pkg")
    table.imports[("pkg.mod", "Alias")] = "pkg.other.Service"
    imported_method = _fn(
        da,
        name="method",
        qual="pkg.other.Service.method",
        path=Path("pkg/other.py"),
    )
    by_qual[imported_method.qual] = imported_method
    resolved_imported = da._resolve_callee(
        "Alias.method",
        caller,
        by_name,
        by_qual,
        symbol_table=table,
        project_root=None,
    )
    assert resolved_imported is imported_method


def test_resolve_callee_class_hierarchy() -> None:
    da = _load()
    path = Path("pkg/mod.py")
    caller = _fn(da, name="caller", qual="pkg.mod.caller", path=path)
    base_method = _fn(
        da,
        name="run",
        qual="pkg.Base.run",
        path=Path("pkg/base.py"),
    )
    by_name = {"run": [base_method]}
    by_qual = {base_method.qual: base_method}
    class_index = {
        "pkg.Base": da.ClassInfo(qual="pkg.Base", module="pkg", bases=[], methods={"run"}),
        "pkg.Service": da.ClassInfo(qual="pkg.Service", module="pkg", bases=["Base"], methods=set()),
    }
    table = da.SymbolTable()
    table.internal_roots.add("pkg")
    table.imports[("pkg.mod", "Service")] = "pkg.Service"
    resolved = da._resolve_callee(
        "Service.run",
        caller,
        by_name,
        by_qual,
        symbol_table=table,
        project_root=None,
        class_index=class_index,
    )
    assert resolved is base_method


def test_resolve_callee_self_class_candidates() -> None:
    da = _load()
    path = Path("pkg/mod.py")
    caller = _fn(
        da,
        name="caller",
        qual="pkg.mod.Service.caller",
        path=path,
        class_name="Service",
    )
    method = _fn(
        da,
        name="run",
        qual="pkg.mod.Service.run",
        path=path,
        class_name="Service",
    )
    by_name = {"run": [method]}
    by_qual = {method.qual: method}
    class_index = {
        "pkg.mod.Service": da.ClassInfo(
            qual="pkg.mod.Service",
            module="pkg.mod",
            bases=[],
            methods={"run"},
        )
    }
    resolved = da._resolve_callee(
        "self.run",
        caller,
        by_name,
        by_qual,
        symbol_table=None,
        project_root=None,
        class_index=class_index,
    )
    assert resolved is method



#!/usr/bin/env python3
"""Pin GitHub Actions to full commit SHAs in workflow files.

Uses the GitHub CLI to resolve tags to commit SHAs.
"""
from __future__ import annotations

import argparse
import json
import re
import subprocess
from dataclasses import dataclass
from pathlib import Path


USES_RE = re.compile(r"^(\s*(?:-\s+)?uses:\s+)([^@\s]+)@([^\s]+)\s*$")
SHA_RE = re.compile(r"^[0-9a-f]{40}$")


@dataclass(frozen=True)
class ActionRef:
    owner: str
    repo: str
    ref: str

    @property
    def full_name(self) -> str:
        return f"{self.owner}/{self.repo}"


def _run_gh(args: list[str]) -> dict:
    result = subprocess.run(
        ["gh", *args],
        check=True,
        capture_output=True,
        text=True,
    )
    return json.loads(result.stdout)


def _resolve_ref(action: ActionRef) -> str:
    if SHA_RE.fullmatch(action.ref):
        return action.ref
    ref = action.ref
    if ref.startswith("v"):
        ref = ref
    payload = _run_gh(
        [
            "api",
            f"repos/{action.owner}/{action.repo}/git/ref/tags/{ref}",
        ]
    )
    obj = payload.get("object", {})
    sha = obj.get("sha")
    if not sha:
        raise SystemExit(f"Unable to resolve tag {action.full_name}@{action.ref}")
    if obj.get("type") == "tag":
        tag_payload = _run_gh(
            [
                "api",
                f"repos/{action.owner}/{action.repo}/git/tags/{sha}",
            ]
        )
        sha = tag_payload.get("object", {}).get("sha", sha)
    if not sha or not SHA_RE.fullmatch(sha):
        raise SystemExit(f"Resolved SHA is invalid for {action.full_name}@{action.ref}")
    return sha


def _parse_action(ref: str) -> ActionRef | None:
    if "@" not in ref:
        return None
    name, version = ref.split("@", 1)
    if "/" not in name:
        return None
    if "<" in version or ">" in version:
        return None
    owner, repo = name.split("/", 1)
    return ActionRef(owner=owner, repo=repo, ref=version)


def _pin_file(path: Path) -> tuple[int, list[str]]:
    lines = path.read_text().splitlines()
    updated = 0
    pinned: list[str] = []
    for i, line in enumerate(lines):
        match = USES_RE.match(line)
        if not match:
            continue
        prefix, name, ref = match.groups()
        action = _parse_action(f"{name}@{ref}")
        if action is None:
            continue
        if SHA_RE.fullmatch(action.ref):
            continue
        sha = _resolve_ref(action)
        lines[i] = f"{prefix}{action.full_name}@{sha}"
        updated += 1
        pinned.append(f"{action.full_name}@{action.ref} -> {sha}")
    if updated:
        path.write_text("\n".join(lines) + "\n")
    return updated, pinned


def main() -> int:
    parser = argparse.ArgumentParser(description="Pin GitHub Actions to SHAs.")
    parser.add_argument(
        "paths",
        nargs="+",
        help="Workflow files to update in-place.",
    )
    args = parser.parse_args()

    total = 0
    for raw in args.paths:
        path = Path(raw)
        if not path.exists():
            raise SystemExit(f"File not found: {path}")
        updated, pinned = _pin_file(path)
        total += updated
        for line in pinned:
            print(f"{path}: {line}")
    print(f"Updated {total} action reference(s).")
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



#!/usr/bin/env python3
"""Thin wrapper for the packaged dataflow audit."""
from __future__ import annotations

from pathlib import Path
import sys


def _bootstrap_src() -> None:
    repo_root = Path(__file__).resolve().parents[1]
    src_path = repo_root / "src"
    sys.path.insert(0, str(src_path))


_bootstrap_src()

from gabion.analysis.dataflow_audit import main  # noqa: E402


if __name__ == "__main__":
    main()



from __future__ import annotations

import argparse
from pathlib import Path
import re
import subprocess
import tomllib


def _run(cmd: list[str]) -> None:
    subprocess.run(cmd, check=True)


def _git_output(*args: str) -> str:
    result = subprocess.run(["git", *args], check=True, stdout=subprocess.PIPE, text=True)
    return result.stdout.strip()


def _read_project_version(pyproject_path: Path) -> str:
    data = tomllib.loads(pyproject_path.read_text(encoding="utf-8"))
    version = data.get("project", {}).get("version")
    if not version:
        raise SystemExit("project.version not found in pyproject.toml")
    return version


def _read_package_version(init_path: Path) -> str:
    text = init_path.read_text(encoding="utf-8")
    match = re.search(r'^__version__\s*=\s*"([^"]+)"\s*$', text, re.M)
    if not match:
        raise SystemExit("__version__ not found in src/gabion/__init__.py")
    return match.group(1)


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("--tag", required=True)
    args = parser.parse_args()
    tag = args.tag

    if not tag.startswith("v"):
        raise SystemExit(f"Unexpected tag format: {tag}")

    _run(["git", "fetch", "origin", "main", "next", "release", "--tags"])

    tag_sha = _git_output("rev-parse", f"refs/tags/{tag}^{{commit}}")
    main_sha = _git_output("rev-parse", "origin/main")
    next_sha = _git_output("rev-parse", "origin/next")
    release_sha = _git_output("rev-parse", "origin/release")

    if tag_sha != main_sha:
        raise SystemExit(f"Release tag must point at main.\ntag={tag_sha}\nmain={main_sha}")
    if tag_sha != next_sha:
        raise SystemExit(f"Release tag must point at next.\ntag={tag_sha}\nnext={next_sha}")
    if tag_sha != release_sha:
        raise SystemExit(f"Release tag must point at release.\ntag={tag_sha}\nrelease={release_sha}")

    project_version = _read_project_version(Path("pyproject.toml"))
    package_version = _read_package_version(Path("src/gabion/__init__.py"))
    tag_version = tag[1:]

    if tag_version != project_version:
        raise SystemExit(
            "Tag version must match project.version in pyproject.toml.\n"
            f"tag={tag_version}\nproject.version={project_version}"
        )
    if project_version != package_version:
        raise SystemExit(
            "project.version must match src/gabion/__init__.__version__.\n"
            f"project.version={project_version}\n__version__={package_version}"
        )


if __name__ == "__main__":
    main()



from __future__ import annotations

from pathlib import Path
import sys


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis import dataflow_audit as da

    return da


def test_config_bundles_and_documented_markers(tmp_path: Path) -> None:
    da = _load()
    config_path = tmp_path / "config.py"
    config_path.write_text(
        "from dataclasses import dataclass\n"
        "\n"
        "@dataclass\n"
        "class AppConfig:\n"
        "    host: str\n"
        "    port: int\n"
        "\n"
        "@dataclass\n"
        "class Helper:\n"
        "    alpha_fn: str\n"
        "    beta: int\n"
        "\n"
        "class PlainConfig:\n"
        "    enabled: bool\n"
        "    mode_fn: str\n"
    )
    bundles = da._iter_config_fields(config_path)
    assert bundles["AppConfig"] == {"host", "port"}
    assert bundles["Helper"] == {"alpha_fn"}
    assert bundles["PlainConfig"] == {"enabled", "mode_fn"}

    other_path = tmp_path / "other.py"
    other_path.write_text("class Nope:\n    pass\n")
    collected = da._collect_config_bundles([config_path, other_path])
    assert config_path in collected
    assert other_path not in collected

    doc_path = tmp_path / "doc.py"
    doc_path.write_text(
        "# dataflow-bundle: a, b\n"
        "# dataflow-bundle:\n"
        "# dataflow-bundle: single\n"
        "# dataflow-bundle: c d\n"
    )
    documented = da._iter_documented_bundles(doc_path)
    assert ("a", "b") in documented
    assert ("c", "d") in documented
    assert ("single",) not in documented


def test_dataclass_registry_and_call_bundles(tmp_path: Path) -> None:
    da = _load()
    pkg = tmp_path / "pkg"
    pkg.mkdir()
    (pkg / "__init__.py").write_text(
        "from .types import ExternalConfig\n"
        "__all__ = [\"ExternalConfig\"]\n"
    )
    (pkg / "types.py").write_text(
        "from dataclasses import dataclass\n"
        "\n"
        "@dataclass\n"
        "class ExternalConfig:\n"
        "    x: int\n"
        "    y: int\n"
        "\n"
        "@dataclass\n"
        "class LocalOnly:\n"
        "    alpha: str\n"
        "    beta: str\n"
    )
    consumer = pkg / "consumer.py"
    consumer.write_text(
        "from dataclasses import dataclass\n"
        "from pkg.types import ExternalConfig\n"
        "from pkg.types import *\n"
        "import pkg.types as types\n"
        "\n"
        "@dataclass\n"
        "class LocalConfig:\n"
        "    a: int\n"
        "    b: int\n"
        "\n"
        "def build():\n"
        "    ExternalConfig(1, 2)\n"
        "    ExternalConfig(x=1, y=2)\n"
        "    LocalConfig(3, 4)\n"
        "    LocalOnly('a', 'b')\n"
        "    types.ExternalConfig(5, 6)\n"
        "    ExternalConfig(*[1, 2])\n"
        "    ExternalConfig(**{\"x\": 1, \"y\": 2})\n"
    )

    paths = [pkg / "__init__.py", pkg / "types.py", consumer]
    registry = da._collect_dataclass_registry(paths, project_root=tmp_path)
    assert "pkg.types.ExternalConfig" in registry

    symbol_table = da._build_symbol_table(paths, tmp_path, external_filter=False)
    bundles = da._iter_dataclass_call_bundles(
        consumer,
        project_root=tmp_path,
        symbol_table=symbol_table,
        dataclass_registry=registry,
    )
    assert ("x", "y") in bundles
    assert ("a", "b") in bundles
    assert ("alpha", "beta") in bundles


def test_config_and_documented_bundles_error_paths(tmp_path: Path) -> None:
    da = _load()
    missing = tmp_path / "missing.py"
    assert da._iter_documented_bundles(missing) == set()

    bad_config = tmp_path / "bad.py"
    bad_config.write_text("def broken(:\n")
    assert da._iter_config_fields(bad_config) == {}


def test_collect_dataclass_registry_without_project_root(tmp_path: Path) -> None:
    da = _load()
    good = tmp_path / "good.py"
    good.write_text(
        "from dataclasses import dataclass\n"
        "\n"
        "@dataclass\n"
        "class Config:\n"
        "    a: int\n"
        "    b: int\n"
    )
    bad = tmp_path / "bad.py"
    bad.write_text("def broken(:\n")
    registry = da._collect_dataclass_registry([good, bad], project_root=None)
    assert any(key.endswith(".Config") for key in registry)


def test_iter_dataclass_call_bundles_invalid_file(tmp_path: Path) -> None:
    da = _load()
    bad = tmp_path / "bad.py"
    bad.write_text("def broken(:\n")
    bundles = da._iter_dataclass_call_bundles(bad, project_root=tmp_path)
    assert bundles == set()



from __future__ import annotations

from typing import Iterable, List, Set


def _jaccard(left: Set[str], right: Set[str]) -> float:
    if not left and not right:
        return 1.0
    union = left | right
    return len(left & right) / len(union)


def merge_bundles(
    bundles: Iterable[Set[str]],
    min_overlap: float = 0.75,
) -> List[Set[str]]:
    merged: List[Set[str]] = [set(b) for b in bundles]
    changed = True
    while changed:
        changed = False
        merged.sort(key=lambda b: (len(b), sorted(b)))
        result: List[Set[str]] = []
        while merged:
            current = merged.pop(0)
            merged_any = False
            for idx, other in enumerate(list(merged)):
                if _jaccard(current, other) >= min_overlap:
                    current |= other
                    merged.pop(idx)
                    merged_any = True
                    changed = True
                    break
            if merged_any:
                merged.append(current)
            else:
                result.append(current)
        merged = result
    merged.sort(key=lambda b: (len(b), sorted(b)))
    return merged



from __future__ import annotations

from pathlib import Path
import sys
import textwrap


def _load_config_module():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.config import dataflow_defaults, merge_payload

    return dataflow_defaults, merge_payload


def test_dataflow_defaults_reads_toml(tmp_path: Path) -> None:
    config_path = tmp_path / "gabion.toml"
    config_path.write_text(
        textwrap.dedent(
            """
            [dataflow]
            exclude = ["in", "extra"]
            ignore_params = ["self", "cls"]
            strictness = "low"
            allow_external = true
            type_audit = true
            fail_on_type_ambiguities = true
            """
        ).strip()
        + "\n"
    )
    dataflow_defaults, _ = _load_config_module()
    defaults = dataflow_defaults(root=tmp_path, config_path=config_path)
    assert defaults["exclude"] == ["in", "extra"]
    assert defaults["ignore_params"] == ["self", "cls"]
    assert defaults["strictness"] == "low"
    assert defaults["allow_external"] is True
    assert defaults["type_audit"] is True
    assert defaults["fail_on_type_ambiguities"] is True


def test_merge_payload_prefers_explicit_values(tmp_path: Path) -> None:
    dataflow_defaults, merge_payload = _load_config_module()
    defaults = {
        "exclude": ["in"],
        "ignore_params": ["self"],
        "strictness": "high",
        "allow_external": False,
    }
    payload = {
        "exclude": None,
        "ignore_params": ["cls"],
        "strictness": None,
        "allow_external": True,
    }
    merged = merge_payload(payload, defaults)
    assert merged["exclude"] == ["in"]
    assert merged["ignore_params"] == ["cls"]
    assert merged["strictness"] == "high"
    assert merged["allow_external"] is True



from __future__ import annotations

import json
from pathlib import Path
from typing import Callable
from urllib.parse import unquote, urlparse

from pygls.lsp.server import LanguageServer
from lsprotocol.types import (
    TEXT_DOCUMENT_DID_OPEN,
    TEXT_DOCUMENT_DID_SAVE,
    TEXT_DOCUMENT_CODE_ACTION,
    CodeAction,
    CodeActionKind,
    CodeActionParams,
    Command,
    Diagnostic,
    DiagnosticSeverity,
    Position,
    Range,
    WorkspaceEdit,
)

from gabion.analysis import (
    AuditConfig,
    analyze_paths,
    apply_baseline,
    compute_violations,
    build_refactor_plan,
    build_synthesis_plan,
    load_baseline,
    render_dot,
    render_protocol_stubs,
    render_refactor_plan,
    render_report,
    render_synthesis_section,
    resolve_baseline_path,
    write_baseline,
)
from gabion.config import dataflow_defaults, merge_payload
from gabion.refactor import (
    FieldSpec,
    RefactorEngine,
    RefactorRequest as RefactorRequestModel,
)
from gabion.schema import (
    RefactorRequest,
    RefactorResponse,
    SynthesisResponse,
    SynthesisRequest,
    TextEditDTO,
)
from gabion.synthesis import NamingContext, SynthesisConfig, Synthesizer

server = LanguageServer("gabion", "0.1.0")
DATAFLOW_COMMAND = "gabion.dataflowAudit"
SYNTHESIS_COMMAND = "gabion.synthesisPlan"
REFACTOR_COMMAND = "gabion.refactorProtocol"


def _uri_to_path(uri: str) -> Path:
    parsed = urlparse(uri)
    if parsed.scheme == "file":
        return Path(unquote(parsed.path))
    return Path(uri)


def _normalize_transparent_decorators(value: object) -> set[str] | None:
    if value is None:
        return None
    items: list[str] = []
    if isinstance(value, str):
        items = [part.strip() for part in value.split(",") if part.strip()]
    elif isinstance(value, (list, tuple, set)):
        for item in value:
            if isinstance(item, str):
                items.extend([part.strip() for part in item.split(",") if part.strip()])
    if not items:
        return None
    return set(items)


def _diagnostics_for_path(path_str: str, project_root: Path | None) -> list[Diagnostic]:
    result = analyze_paths(
        [Path(path_str)],
        recursive=True,
        type_audit=False,
        type_audit_report=False,
        type_audit_max=0,
        include_constant_smells=False,
        include_unused_arg_smells=False,
        config=AuditConfig(project_root=project_root),
    )
    diagnostics: list[Diagnostic] = []
    for path, bundles in result.groups_by_path.items():
        span_map = result.param_spans_by_path.get(path, {})
        for fn_name, group_list in bundles.items():
            param_spans = span_map.get(fn_name, {})
            for bundle in group_list:
                message = f"Implicit bundle detected: {', '.join(sorted(bundle))}"
                for name in sorted(bundle):
                    span = param_spans.get(name)
                    if span is None:  # pragma: no cover - spans are derived from parsed params
                        start = Position(line=0, character=0)  # pragma: no cover
                        end = Position(line=0, character=1)  # pragma: no cover
                    else:
                        start_line, start_col, end_line, end_col = span
                        start = Position(line=start_line, character=start_col)
                        end = Position(line=end_line, character=end_col)
                    diagnostics.append(
                        Diagnostic(
                            range=Range(start=start, end=end),
                            message=message,
                            severity=DiagnosticSeverity.Information,
                            source="gabion",
                        )
                    )
    return diagnostics


@server.command(DATAFLOW_COMMAND)
def execute_command(ls: LanguageServer, payload: dict | None = None) -> dict:
    if payload is None:
        payload = {}
    root = payload.get("root") or ls.workspace.root_path or "."
    config_path = payload.get("config")
    defaults = dataflow_defaults(
        Path(root), Path(config_path) if config_path else None
    )
    payload = merge_payload(payload, defaults)

    raw_paths = payload.get("paths") or []
    if raw_paths:
        paths = [Path(p) for p in raw_paths]
    else:
        paths = [Path(root)]
    root = payload.get("root") or root
    report_path = payload.get("report")
    dot_path = payload.get("dot")
    fail_on_violations = payload.get("fail_on_violations", False)
    no_recursive = payload.get("no_recursive", False)
    max_components = payload.get("max_components", 10)
    type_audit = payload.get("type_audit", False)
    type_audit_report = payload.get("type_audit_report", False)
    type_audit_max = payload.get("type_audit_max", 50)
    fail_on_type_ambiguities = payload.get("fail_on_type_ambiguities", False)
    exclude_dirs = set(payload.get("exclude", []))
    ignore_params = set(payload.get("ignore_params", []))
    allow_external = payload.get("allow_external", False)
    strictness = payload.get("strictness", "high")
    transparent_decorators = _normalize_transparent_decorators(
        payload.get("transparent_decorators")
    )
    baseline_path = resolve_baseline_path(payload.get("baseline"), Path(root))
    baseline_write = bool(payload.get("baseline_write", False)) and baseline_path is not None
    synthesis_plan_path = payload.get("synthesis_plan")
    synthesis_report = payload.get("synthesis_report", False)
    synthesis_max_tier = payload.get("synthesis_max_tier", 2)
    synthesis_min_bundle_size = payload.get("synthesis_min_bundle_size", 2)
    synthesis_allow_singletons = payload.get("synthesis_allow_singletons", False)
    synthesis_protocols_path = payload.get("synthesis_protocols")
    synthesis_protocols_kind = payload.get("synthesis_protocols_kind", "dataclass")
    refactor_plan = payload.get("refactor_plan", False)
    refactor_plan_json = payload.get("refactor_plan_json")

    config = AuditConfig(
        project_root=Path(root),
        exclude_dirs=exclude_dirs,
        ignore_params=ignore_params,
        external_filter=not allow_external,
        strictness=strictness,
        transparent_decorators=transparent_decorators,
    )
    if fail_on_type_ambiguities:
        type_audit = True
    analysis = analyze_paths(
        paths,
        recursive=not no_recursive,
        type_audit=type_audit or type_audit_report,
        type_audit_report=type_audit_report,
        type_audit_max=type_audit_max,
        include_constant_smells=bool(report_path),
        include_unused_arg_smells=bool(report_path),
        config=config,
    )

    response: dict = {
        "type_suggestions": analysis.type_suggestions,
        "type_ambiguities": analysis.type_ambiguities,
        "unused_arg_smells": analysis.unused_arg_smells,
    }

    synthesis_plan: dict[str, object] | None = None
    if synthesis_plan_path or synthesis_report or synthesis_protocols_path:
        synthesis_plan = build_synthesis_plan(
            analysis.groups_by_path,
            project_root=Path(root),
            max_tier=int(synthesis_max_tier),
            min_bundle_size=int(synthesis_min_bundle_size),
            allow_singletons=bool(synthesis_allow_singletons),
            config=config,
        )
        if synthesis_plan_path:
            payload_json = json.dumps(synthesis_plan, indent=2, sort_keys=True)
            if synthesis_plan_path == "-":
                response["synthesis_plan"] = synthesis_plan
            else:
                Path(synthesis_plan_path).write_text(payload_json)
        if synthesis_protocols_path:
            stubs = render_protocol_stubs(
                synthesis_plan, kind=str(synthesis_protocols_kind)
            )
            if synthesis_protocols_path == "-":
                response["synthesis_protocols"] = stubs
            else:
                Path(synthesis_protocols_path).write_text(stubs)

    refactor_plan_payload: dict[str, object] | None = None
    if refactor_plan or refactor_plan_json:
        refactor_plan_payload = build_refactor_plan(
            analysis.groups_by_path,
            paths,
            config=config,
        )
        if refactor_plan_json:
            payload_json = json.dumps(refactor_plan_payload, indent=2, sort_keys=True)
            if refactor_plan_json == "-":
                response["refactor_plan"] = refactor_plan_payload
            else:
                Path(refactor_plan_json).write_text(payload_json)

    if dot_path:
        dot = render_dot(analysis.groups_by_path)
        if dot_path == "-":
            response["dot"] = dot
        else:
            Path(dot_path).write_text(dot)

    violations: list[str] = []
    effective_violations: list[str] | None = None
    if report_path:
        report, violations = render_report(
            analysis.groups_by_path,
            max_components,
            type_suggestions=analysis.type_suggestions if type_audit_report else None,
            type_ambiguities=analysis.type_ambiguities if type_audit_report else None,
            constant_smells=analysis.constant_smells,
            unused_arg_smells=analysis.unused_arg_smells,
        )
        if baseline_path is not None:
            baseline_entries = load_baseline(baseline_path)
            if baseline_write:
                write_baseline(baseline_path, violations)
                baseline_entries = set(violations)
                effective_violations = []
            else:
                effective_violations, _ = apply_baseline(violations, baseline_entries)
            report = (
                report
                + "\n\nBaseline/Ratchet:\n```\n"
                + f"Baseline: {baseline_path}\n"
                + f"Baseline entries: {len(baseline_entries)}\n"
                + f"New violations: {len(effective_violations)}\n"
                + "```\n"
            )
        if synthesis_plan and (
            synthesis_report or synthesis_plan_path or synthesis_protocols_path
        ):
            report = report + render_synthesis_section(synthesis_plan)
        if refactor_plan_payload and (refactor_plan or refactor_plan_json):
            report = report + render_refactor_plan(refactor_plan_payload)
        Path(report_path).write_text(report)
    else:
        violations = compute_violations(
            analysis.groups_by_path,
            max_components,
            type_suggestions=analysis.type_suggestions if type_audit_report else None,
            type_ambiguities=analysis.type_ambiguities if type_audit_report else None,
        )
        if baseline_path is not None:
            baseline_entries = load_baseline(baseline_path)
            if baseline_write:
                write_baseline(baseline_path, violations)
                effective_violations = []
            else:
                effective_violations, _ = apply_baseline(violations, baseline_entries)

    if effective_violations is None:
        effective_violations = violations
    response["violations"] = len(effective_violations)
    if baseline_path is not None:
        response["baseline_path"] = str(baseline_path)
        response["baseline_written"] = bool(baseline_write)
    if fail_on_type_ambiguities and analysis.type_ambiguities:
        response["exit_code"] = 1
    else:
        if baseline_write:
            response["exit_code"] = 0
        else:
            response["exit_code"] = 1 if (fail_on_violations and effective_violations) else 0
    return response


@server.command(SYNTHESIS_COMMAND)
def execute_synthesis(ls: LanguageServer, payload: dict | None = None) -> dict:
    if payload is None:
        payload = {}
    try:
        request = SynthesisRequest.model_validate(payload)
    except Exception as exc:  # pydantic validation
        return {"protocols": [], "warnings": [], "errors": [str(exc)]}

    bundle_tiers: dict[frozenset[str], int] = {}
    for entry in request.bundles:
        bundle = entry.bundle
        if not bundle:
            continue
        bundle_tiers[frozenset(bundle)] = entry.tier

    field_types = request.field_types or {}
    config = SynthesisConfig(
        max_tier=request.max_tier,
        min_bundle_size=request.min_bundle_size,
        allow_singletons=request.allow_singletons,
        merge_overlap_threshold=request.merge_overlap_threshold,
    )
    naming_context = NamingContext(
        existing_names=set(request.existing_names),
        frequency=request.frequency or {},
        fallback_prefix=request.fallback_prefix,
    )
    plan = Synthesizer(config=config).plan(
        bundle_tiers=bundle_tiers,
        field_types=field_types,
        naming_context=naming_context,
    )
    response = SynthesisResponse(
        protocols=[
            {
                "name": spec.name,
                "fields": [
                    {
                        "name": field.name,
                        "type_hint": field.type_hint,
                        "source_params": sorted(field.source_params),
                    }
                    for field in spec.fields
                ],
                "bundle": sorted(spec.bundle),
                "tier": spec.tier,
                "rationale": spec.rationale,
            }
            for spec in plan.protocols
        ],
        warnings=plan.warnings,
        errors=plan.errors,
    )
    return response.model_dump()


@server.command(REFACTOR_COMMAND)
def execute_refactor(ls: LanguageServer, payload: dict | None = None) -> dict:
    if payload is None:
        payload = {}
    try:
        request = RefactorRequest.model_validate(payload)
    except Exception as exc:  # pydantic validation
        return RefactorResponse(errors=[str(exc)]).model_dump()

    project_root = None
    if ls.workspace.root_path:
        project_root = Path(ls.workspace.root_path)
    engine = RefactorEngine(project_root=project_root)
    plan = engine.plan_protocol_extraction(
        RefactorRequestModel(
            protocol_name=request.protocol_name,
            bundle=request.bundle,
            fields=[
                FieldSpec(name=field.name, type_hint=field.type_hint)
                for field in request.fields or []
            ],
            target_path=request.target_path,
            target_functions=request.target_functions,
            compatibility_shim=request.compatibility_shim,
            rationale=request.rationale,
        )
    )
    edits = [
        TextEditDTO(
            path=edit.path,
            start=edit.start,
            end=edit.end,
            replacement=edit.replacement,
        )
        for edit in plan.edits
    ]
    response = RefactorResponse(
        edits=edits,
        warnings=plan.warnings,
        errors=plan.errors,
    )
    return response.model_dump()


@server.feature(TEXT_DOCUMENT_CODE_ACTION)
def code_action(ls: LanguageServer, params: CodeActionParams) -> list[CodeAction]:
    path = _uri_to_path(params.text_document.uri)
    payload = {
        "protocol_name": "TODO_Bundle",
        "bundle": [],
        "target_path": str(path),
        "target_functions": [],
        "rationale": "Stub code action; populate bundle details manually.",
    }
    title = "Gabion: Extract Protocol (stub)"
    return [
        CodeAction(
            title=title,
            kind=CodeActionKind.RefactorExtract,
            command=Command(title=title, command=REFACTOR_COMMAND, arguments=[payload]),
            edit=WorkspaceEdit(changes={}),
        )
    ]


@server.feature(TEXT_DOCUMENT_DID_OPEN)
def did_open(ls: LanguageServer, params) -> None:
    uri = params.text_document.uri
    doc = ls.workspace.get_document(uri)
    root = Path(ls.workspace.root_path) if ls.workspace.root_path else None
    diagnostics = _diagnostics_for_path(doc.path, root)
    ls.publish_diagnostics(uri, diagnostics)


@server.feature(TEXT_DOCUMENT_DID_SAVE)
def did_save(ls: LanguageServer, params) -> None:
    uri = params.text_document.uri
    doc = ls.workspace.get_document(uri)
    root = Path(ls.workspace.root_path) if ls.workspace.root_path else None
    diagnostics = _diagnostics_for_path(doc.path, root)
    ls.publish_diagnostics(uri, diagnostics)


def start(start_fn: Callable[[], None] | None = None) -> None:
    """Start the language server (stub)."""
    (start_fn or server.start_io)()


if __name__ == "__main__":  # pragma: no cover
    start()  # pragma: no cover



from __future__ import annotations

from pathlib import Path
import sys


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis import dataflow_audit

    return dataflow_audit


def _write_sample_code(path: Path) -> None:
    path.write_text(
        "def helper(x, y):\n"
        "    return x + y\n\n"
        "def alpha(a, b, *args, **kwargs):\n"
        "    c = a\n"
        "    d, e = (b, b)\n"
        "    return helper(c, d)\n"
    )


def test_run_generates_outputs(tmp_path: Path) -> None:
    dataflow_audit = _load()
    sample = tmp_path / "sample.py"
    _write_sample_code(sample)
    report = tmp_path / "report.md"
    dot = tmp_path / "graph.dot"
    plan = tmp_path / "plan.json"
    protocols = tmp_path / "protocols.py"
    refactor = tmp_path / "refactor.json"
    argv = [
        str(tmp_path),
        "--root",
        str(tmp_path),
        "--report",
        str(report),
        "--dot",
        str(dot),
        "--synthesis-plan",
        str(plan),
        "--synthesis-protocols",
        str(protocols),
        "--synthesis-protocols-kind",
        "protocol",
        "--synthesis-min-bundle-size",
        "1",
        "--synthesis-allow-singletons",
        "--refactor-plan",
        "--refactor-plan-json",
        str(refactor),
        "--type-audit-report",
        "--type-audit-max",
        "5",
        "--transparent-decorators",
        "decorator_a,decorator_b",
    ]
    code = dataflow_audit.run(argv)
    assert code == 0
    assert report.exists()
    assert dot.exists()
    assert plan.exists()
    assert protocols.exists()
    assert refactor.exists()


def test_run_baseline_write_and_apply(tmp_path: Path) -> None:
    dataflow_audit = _load()
    sample = tmp_path / "sample.py"
    _write_sample_code(sample)
    baseline = tmp_path / "baseline.txt"
    report = tmp_path / "report.md"
    code = dataflow_audit.run(
        [
            str(tmp_path),
            "--root",
            str(tmp_path),
            "--baseline",
            str(baseline),
            "--baseline-write",
            "--report",
            str(report),
            "--fail-on-violations",
        ]
    )
    assert code == 0
    assert baseline.exists()

    code = dataflow_audit.run(
        [
            str(tmp_path),
            "--root",
            str(tmp_path),
            "--baseline",
            str(baseline),
            "--report",
            str(report),
            "--fail-on-violations",
        ]
    )
    assert code == 0


def test_run_invalid_strictness_from_config(tmp_path: Path) -> None:
    dataflow_audit = _load()
    sample = tmp_path / "sample.py"
    _write_sample_code(sample)
    config = tmp_path / "gabion.toml"
    config.write_text("[dataflow]\nstrictness = \"weird\"\n")
    code = dataflow_audit.run(
        [
            str(tmp_path),
            "--root",
            str(tmp_path),
            "--config",
            str(config),
            "--report",
            str(tmp_path / "report.md"),
        ]
    )
    assert code == 0



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from gabion import server


class _DummyWorkspace:
    def __init__(self, root_path: str) -> None:
        self.root_path = root_path


class _DummyServer:
    def __init__(self, root_path: str) -> None:
        self.workspace = _DummyWorkspace(root_path)


@dataclass
class _CommandResult:
    exit_code: int
    violations: int


def test_normalize_transparent_decorators() -> None:
    assert server._normalize_transparent_decorators(None) is None
    assert server._normalize_transparent_decorators("a, b") == {"a", "b"}
    assert server._normalize_transparent_decorators(["a", "b,c"]) == {"a", "b", "c"}
    assert server._normalize_transparent_decorators([]) is None


def test_uri_to_path_file_scheme() -> None:
    path = server._uri_to_path("file:///tmp/example.py")
    assert str(path).endswith("/tmp/example.py")


def test_uri_to_path_plain_path() -> None:
    path = server._uri_to_path("/tmp/example.py")
    assert str(path).endswith("/tmp/example.py")


def _write_minimal_module(path: Path) -> None:
    path.write_text(
        "def alpha():\n"
        "    return 1\n"
    )


def test_execute_command_no_violations(tmp_path: Path) -> None:
    module_path = tmp_path / "sample.py"
    _write_minimal_module(module_path)
    ls = _DummyServer(str(tmp_path))
    result = server.execute_command(
        ls,
        {
            "root": str(tmp_path),
            "paths": [str(module_path)],
            "fail_on_violations": True,
            "type_audit": False,
        },
    )
    assert _CommandResult(
        exit_code=result.get("exit_code", -1),
        violations=result.get("violations", -1),
    ) == _CommandResult(exit_code=0, violations=0)


def test_execute_command_baseline_write(tmp_path: Path) -> None:
    module_path = tmp_path / "sample.py"
    _write_minimal_module(module_path)
    baseline_path = tmp_path / "baseline.txt"
    ls = _DummyServer(str(tmp_path))
    result = server.execute_command(
        ls,
        {
            "root": str(tmp_path),
            "paths": [str(module_path)],
            "baseline": str(baseline_path),
            "baseline_write": True,
        },
    )
    assert result.get("baseline_written") is True
    assert baseline_path.exists()
    baseline_text = baseline_path.read_text()
    assert baseline_text.startswith("# gabion baseline")


def test_execute_command_report_and_dot(tmp_path: Path) -> None:
    module_path = tmp_path / "sample.py"
    _write_minimal_module(module_path)
    report_path = tmp_path / "report.md"
    dot_path = tmp_path / "graph.dot"
    ls = _DummyServer(str(tmp_path))
    result = server.execute_command(
        ls,
        {
            "root": str(tmp_path),
            "paths": [str(module_path)],
            "report": str(report_path),
            "dot": str(dot_path),
        },
    )
    assert report_path.exists()
    assert dot_path.exists()
    assert "violations" in result


def test_execute_command_synthesis_outputs(tmp_path: Path) -> None:
    module_path = tmp_path / "sample.py"
    _write_minimal_module(module_path)
    plan_path = tmp_path / "plan.json"
    protocol_path = tmp_path / "protocols.py"
    ls = _DummyServer(str(tmp_path))
    result = server.execute_command(
        ls,
        {
            "root": str(tmp_path),
            "paths": [str(module_path)],
            "synthesis_plan": str(plan_path),
            "synthesis_protocols": str(protocol_path),
        },
    )
    assert plan_path.exists()
    assert protocol_path.exists()
    assert "violations" in result


def test_execute_synthesis_minimal_payload() -> None:
    result = server.execute_synthesis(None, {"bundles": []})
    assert result["protocols"] == []
    assert result["errors"] == []


def test_execute_refactor_invalid_payload() -> None:
    result = server.execute_refactor(None, {})
    assert result["errors"]



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Optional, Set, Tuple


@dataclass
class SymbolTable:
    imports: Dict[Tuple[str, str], str] = field(default_factory=dict)
    # Map: (module_name, local_name) -> fully_qualified_name


@dataclass
class ClassInfo:
    qual: str
    bases: List[str]
    methods: Set[str]


@dataclass
class DispatchTable:
    name: str
    targets: Set[str]


@dataclass(frozen=True)
class CallArgs:
    callee_expr: str
    pos_map: Dict[str, str]
    kw_map: Dict[str, str]
    resolved_targets: List[str] = field(default_factory=list)


@dataclass
class FunctionInfo:
    qual: str
    params: List[str]
    calls: List[CallArgs]


@dataclass
class ParamUse:
    direct_forward: Set[Tuple[str, str]]
    non_forward: bool
    current_aliases: Set[str]



from __future__ import annotations

from pathlib import Path
import os
import textwrap

import libcst as cst

from gabion.refactor.engine import (
    RefactorEngine,
    _CallSiteTransformer,
    _RefactorTransformer,
    _collect_import_context,
    _has_typing_overload_import,
    _has_typing_protocol_import,
    _module_expr_to_str,
    _module_name,
    _rewrite_call_sites,
    _rewrite_call_sites_in_project,
)
from gabion.refactor.model import FieldSpec, RefactorRequest


def test_plan_protocol_extraction_relative_path_and_fields(tmp_path: Path) -> None:
    target = tmp_path / "mod.py"
    target.write_text("def f(a, b):\n    return a\n")
    request = RefactorRequest(
        protocol_name="Bundle",
        bundle=[],
        fields=[FieldSpec(name="a"), FieldSpec(name=""), FieldSpec(name="a"), FieldSpec(name="b")],
        target_path="mod.py",
    )
    plan = RefactorEngine(project_root=tmp_path).plan_protocol_extraction(request)
    assert plan.errors == []
    assert plan.edits


def test_plan_protocol_extraction_typing_import_variants(tmp_path: Path) -> None:
    engine = RefactorEngine(project_root=tmp_path)

    mod_typing = tmp_path / "typing_mod.py"
    mod_typing.write_text("import typing\n\ndef f(a):\n    return a\n")
    plan_typing = engine.plan_protocol_extraction(
        RefactorRequest(
            protocol_name="Proto",
            bundle=["a"],
            target_path=str(mod_typing),
        )
    )
    assert plan_typing.edits
    assert "typing.Protocol" in plan_typing.edits[0].replacement

    mod_protocol = tmp_path / "protocol_mod.py"
    mod_protocol.write_text("from typing import Protocol\n\ndef f(a):\n    return a\n")
    plan_protocol = engine.plan_protocol_extraction(
        RefactorRequest(
            protocol_name="Proto",
            bundle=["a"],
            target_path=str(mod_protocol),
        )
    )
    assert plan_protocol.edits
    assert "class Proto(Protocol)" in plan_protocol.edits[0].replacement


def test_module_name_handles_value_error(tmp_path: Path) -> None:
    assert _module_name(Path("mod.py"), tmp_path / "other") == "mod"


def test_typing_import_helpers_negative_branches() -> None:
    module = cst.parse_module("from typing_extensions import Protocol\n")
    assert _has_typing_protocol_import(list(module.body)) is False
    module = cst.parse_module("from other import overload\n")
    assert _has_typing_overload_import(list(module.body)) is False
    expr = cst.Attribute(cst.Name("pkg"), cst.Name("mod"))
    assert _module_expr_to_str(expr) == "pkg.mod"


def test_rewrite_call_sites_uses_protocol_alias(tmp_path: Path) -> None:
    source = (
        "from pkg.mod import target, Bundle as PB\n"
        "\n"
        "def f(a, b):\n"
        "    target(a, b)\n"
    )
    module = cst.parse_module(source)
    warnings, updated = _rewrite_call_sites(
        module,
        file_path=tmp_path / "consumer.py",
        target_path=tmp_path / "target.py",
        target_module="pkg.mod",
        protocol_name="Bundle",
        bundle_fields=["a", "b"],
        targets={"target"},
    )
    assert warnings == []
    assert updated is not None
    assert "PB(" in updated.code


def test_collect_import_context_skips_nonmatching() -> None:
    module = cst.Module(
        body=[
            cst.SimpleStatementLine(
                [cst.Import(names=[cst.ImportAlias(name=cst.Name("other"))])]
            ),
            cst.SimpleStatementLine(
                [
                    cst.ImportFrom(
                        module=cst.Name("pkg"),
                        names=[cst.ImportStar()],
                    )
                ]
            ),
            cst.SimpleStatementLine(
                [
                    cst.ImportFrom(
                        module=cst.Attribute(cst.Name("pkg"), cst.Name("mod")),
                        names=[
                            cst.ImportAlias(
                                name=cst.Attribute(cst.Name("pkg"), cst.Name("Thing"))
                            )
                        ],
                    )
                ]
            ),
        ]
    )
    module_aliases, imported_targets, protocol_alias = _collect_import_context(
        module, target_module="pkg.mod", protocol_name="Protocol"
    )
    assert module_aliases == {}
    assert imported_targets == {}
    assert protocol_alias is None


def test_collect_import_context_skips_import_star_matching_module() -> None:
    module = cst.Module(
        body=[
            cst.SimpleStatementLine(
                [
                    cst.ImportFrom(
                        module=cst.Name("pkg"),
                        names=[cst.ImportStar()],
                    )
                ]
            )
        ]
    )
    module_aliases, imported_targets, protocol_alias = _collect_import_context(
        module, target_module="pkg", protocol_name="Protocol"
    )
    assert module_aliases == {}
    assert imported_targets == {}
    assert protocol_alias is None


def test_rewrite_call_sites_empty_targets(tmp_path: Path) -> None:
    module = cst.parse_module("def f(a):\n    return a\n")
    warnings, updated = _rewrite_call_sites(
        module,
        file_path=tmp_path / "a.py",
        target_path=tmp_path / "a.py",
        target_module="pkg.mod",
        protocol_name="Bundle",
        bundle_fields=["a"],
        targets=set(),
    )
    assert warnings == []
    assert updated is None


def test_rewrite_call_sites_module_alias_and_method_target(tmp_path: Path) -> None:
    source = (
        "import pkg.mod as pm\n"
        "class C:\n"
        "    def m(self, a, b):\n"
        "        return a\n"
        "    def caller(self, a, b):\n"
        "        return self.m(a, b)\n"
        "\n"
        "def f(a, b):\n"
        "    return pm.target(a, b)\n"
    )
    module = cst.parse_module(source)
    warnings, updated = _rewrite_call_sites(
        module,
        file_path=tmp_path / "consumer.py",
        target_path=tmp_path / "target.py",
        target_module="pkg.mod",
        protocol_name="Bundle",
        bundle_fields=["a", "b"],
        targets={"target", "C.m"},
    )
    assert updated is not None
    assert "pm.Bundle" in updated.code
    assert warnings == []


def test_rewrite_call_sites_in_project_read_errors(tmp_path: Path) -> None:
    root = tmp_path / "src"
    root.mkdir()
    target = root / "target.py"
    target.write_text("def target(a, b):\n    return a + b\n")
    unreadable = root / "unreadable.py"
    unreadable.write_text("def nope():\n    return 1\n")
    os.chmod(unreadable, 0)
    consumer = root / "consumer.py"
    consumer.write_text("def f(a):\n    return a\n")
    edits, warnings = _rewrite_call_sites_in_project(
        project_root=tmp_path,
        target_path=target,
        target_module="target",
        protocol_name="Bundle",
        bundle_fields=["a", "b"],
        targets={"target"},
    )
    os.chmod(unreadable, 0o644)
    assert warnings
    assert edits == []


def test_refactor_transformer_helpers() -> None:
    transformer = _RefactorTransformer(
        targets={"f"},
        bundle_fields=["a"],
        protocol_hint="Bundle[",
    )
    params = cst.parse_module("def f(self, a, /, b, *, c):\n    pass\n").body[0].params
    names = transformer._ordered_param_names(params)
    assert names == ["self", "a", "b", "c"]
    self_param = transformer._find_self_param(params, "self")
    assert self_param is not None
    assert transformer._find_self_param(params, "cls") is None
    assert transformer._choose_bundle_name(["bundle", "bundle_1"]) == "bundle_2"
    built = transformer._build_parameters(self_param, "bundle")
    assert built.params[0].name.value == "self"
    assert transformer.warnings
    shim_params = transformer._build_shim_parameters(self_param)
    assert shim_params.params[0].name.value == "self"

    suite = cst.IndentedBlock(
        body=[cst.SimpleStatementLine([cst.Expr(cst.SimpleString('"""Doc"""'))])]
    )
    injected = transformer._inject_preamble(suite, "bundle", ["a"])
    assert isinstance(injected, cst.IndentedBlock)

    passthrough = transformer._inject_preamble(suite, "bundle", [])
    assert passthrough is suite
    passthrough = transformer._inject_preamble(cst.SimpleStatementLine([]), "bundle", ["a"])
    assert passthrough is not None


def test_refactor_transformer_async_and_no_params(tmp_path: Path) -> None:
    target = tmp_path / "mod.py"
    target.write_text(
        textwrap.dedent(
            """
            async def f():
                return 1

            class C:
                async def m(self, a):
                    return a
            """
        ).strip()
        + "\n"
    )
    plan = RefactorEngine(project_root=tmp_path).plan_protocol_extraction(
        RefactorRequest(
            protocol_name="Bundle",
            bundle=["a"],
            target_path=str(target),
            target_functions=["f", "C.m"],
        )
    )
    assert plan.edits


def test_call_site_transformer_helpers() -> None:
    transformer = _CallSiteTransformer(
        file_is_target=True,
        target_simple={"target"},
        target_methods={"C": {"m"}},
        module_aliases=set(),
        imported_targets=set(),
        bundle_fields=["a", "b"],
        constructor_expr=cst.Attribute(cst.Name("mod"), cst.Name("Bundle")),
    )
    transformer._class_stack.append("C")
    call = cst.parse_expression("self.m(a, b)")
    assert transformer._is_target_call(call.func) is True
    transformer._class_stack.pop()

    transformer = _CallSiteTransformer(
        file_is_target=False,
        target_simple={"target"},
        target_methods={},
        module_aliases={"mod"},
        imported_targets=set(),
        bundle_fields=["a", "b"],
        constructor_expr=cst.Attribute(cst.Name("mod"), cst.Name("Bundle")),
    )
    call = cst.parse_expression("mod.target(a, b)")
    assert transformer._is_target_call(call.func) is True

    wrapped_call = cst.parse_expression("target(mod.Bundle(a=a, b=b))")
    assert transformer._already_wrapped(wrapped_call) is True

    name_constructor = _CallSiteTransformer(
        file_is_target=True,
        target_simple={"target"},
        target_methods={},
        module_aliases=set(),
        imported_targets=set(),
        bundle_fields=["a", "b"],
        constructor_expr=cst.Name("Bundle"),
    )
    mismatch_call = cst.parse_expression("target(mod.Bundle(a=a, b=b))")
    assert name_constructor._already_wrapped(mismatch_call) is False



from __future__ import annotations

from pathlib import Path
import sys


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import AuditConfig, build_synthesis_plan

    return AuditConfig, build_synthesis_plan


def test_build_synthesis_plan_merges_overlapping_bundles(tmp_path: Path) -> None:
    AuditConfig, build_synthesis_plan = _load()
    groups_by_path = {
        tmp_path / "a.py": {"f": [{"a", "b", "c"}]},
        tmp_path / "b.py": {"g": [{"a", "b", "c", "d"}]},
    }
    plan = build_synthesis_plan(
        groups_by_path,
        project_root=tmp_path,
        max_tier=3,
        config=AuditConfig(project_root=tmp_path),
    )
    protocols = plan.get("protocols", [])
    bundles = {tuple(sorted(entry.get("bundle", []))) for entry in protocols}
    assert ("a", "b", "c", "d") in bundles
    assert ("a", "b", "c") not in bundles



from __future__ import annotations

from pathlib import Path
import sys


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis import dataflow_audit as da

    return da


def _write_bundle_module(path: Path) -> None:
    path.write_text(
        "def callee(x):\n"
        "    return x\n"
        "\n"
        "def caller(a, b):\n"
        "    callee(a)\n"
        "    callee(b)\n"
    )


def _write_type_module(path: Path) -> None:
    path.write_text(
        "def typed_int(x: int):\n"
        "    return x\n"
        "\n"
        "def typed_str(x: str):\n"
        "    return x\n"
        "\n"
        "def type_caller(a):\n"
        "    typed_int(a)\n"
        "\n"
        "def type_conflict(b):\n"
        "    typed_int(b)\n"
        "    typed_str(b)\n"
    )


def test_run_baseline_write_requires_path(tmp_path: Path, capsys) -> None:
    da = _load()
    target = tmp_path / "mod.py"
    _write_bundle_module(target)
    code = da.run([str(target), "--baseline-write"])
    assert code == 2
    captured = capsys.readouterr()
    assert "Baseline path required" in captured.err


def test_run_dot_stdout_short_circuit(tmp_path: Path, capsys) -> None:
    da = _load()
    target = tmp_path / "mod.py"
    _write_bundle_module(target)
    code = da.run([str(target), "--dot", "-"])
    assert code == 0
    captured = capsys.readouterr()
    assert "digraph dataflow_grammar" in captured.out


def test_run_report_baseline_write_and_apply(tmp_path: Path) -> None:
    da = _load()
    target = tmp_path / "mod.py"
    _write_bundle_module(target)
    baseline = tmp_path / "baseline.txt"
    report = tmp_path / "report.md"
    code = da.run(
        [
            str(target),
            "--report",
            str(report),
            "--baseline",
            str(baseline),
            "--baseline-write",
            "--fail-on-violations",
        ]
    )
    assert code == 0
    assert baseline.exists()
    assert report.exists()

    code = da.run(
        [
            str(target),
            "--report",
            str(report),
            "--baseline",
            str(baseline),
            "--fail-on-violations",
        ]
    )
    assert code == 0


def test_run_fail_on_violations_no_report(tmp_path: Path) -> None:
    da = _load()
    target = tmp_path / "mod.py"
    _write_bundle_module(target)
    code = da.run([str(target), "--fail-on-violations"])
    assert code == 1


def test_run_type_audit_early_return(tmp_path: Path, capsys) -> None:
    da = _load()
    target = tmp_path / "types.py"
    _write_type_module(target)
    code = da.run([str(target), "--type-audit", "--type-audit-max", "1"])
    assert code == 0
    captured = capsys.readouterr()
    assert "Type tightening candidates" in captured.out



from __future__ import annotations

import importlib.util
from pathlib import Path
import sys

import pytest


def _has_pygls() -> bool:
    return importlib.util.find_spec("pygls") is not None


@pytest.mark.skipif(not _has_pygls(), reason="pygls not installed")
def test_run_command_unknown_command_raises() -> None:
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.lsp_client import CommandRequest, LspClientError, run_command

    with pytest.raises(LspClientError):
        run_command(CommandRequest("gabion.unknown", []), root=repo_root)



from __future__ import annotations

import importlib.util
from pathlib import Path
import sys

import pytest


def _has_pygls() -> bool:
    return importlib.util.find_spec("pygls") is not None


@pytest.mark.skipif(not _has_pygls(), reason="pygls not installed")
def test_lsp_execute_command() -> None:
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.lsp_client import CommandRequest, run_command

    result = run_command(
        CommandRequest(
            "gabion.dataflowAudit",
            [{"paths": [str(repo_root)], "fail_on_violations": False}],
        ),
        root=repo_root,
    )
    assert "exit_code" in result
    synth_result = run_command(
        CommandRequest(
            "gabion.synthesisPlan",
            [
                {
                    "bundles": [{"bundle": ["ctx"], "tier": 2}],
                    "min_bundle_size": 1,
                    "allow_singletons": True,
                    "existing_names": ["CtxBundle"],
                }
            ],
        ),
        root=repo_root,
    )
    assert "protocols" in synth_result



from __future__ import annotations

from pathlib import Path
import sys
import textwrap

import libcst as cst


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.refactor.engine import (
        RefactorEngine,
        _find_import_insert_index,
        _module_expr_to_str,
        _module_name,
        _has_typing_import,
    )
    from gabion.refactor.model import FieldSpec, RefactorRequest

    return (
        RefactorEngine,
        FieldSpec,
        RefactorRequest,
        _find_import_insert_index,
        _module_expr_to_str,
        _module_name,
        _has_typing_import,
    )


def test_refactor_engine_handles_missing_file(tmp_path: Path) -> None:
    RefactorEngine, _, RefactorRequest, *_ = _load()
    request = RefactorRequest(
        protocol_name="BundleProtocol",
        bundle=["a"],
        target_path=str(tmp_path / "missing.py"),
    )
    plan = RefactorEngine(project_root=tmp_path).plan_protocol_extraction(request)
    assert plan.errors


def test_refactor_engine_handles_parse_error(tmp_path: Path) -> None:
    RefactorEngine, _, RefactorRequest, *_ = _load()
    target = tmp_path / "bad.py"
    target.write_text("def broken(")
    request = RefactorRequest(
        protocol_name="BundleProtocol",
        bundle=["a"],
        target_path=str(target),
    )
    plan = RefactorEngine(project_root=tmp_path).plan_protocol_extraction(request)
    assert plan.errors


def test_refactor_engine_requires_protocol_name(tmp_path: Path) -> None:
    RefactorEngine, _, RefactorRequest, *_ = _load()
    target = tmp_path / "sample.py"
    target.write_text("def f(a):\n    return a\n")
    request = RefactorRequest(
        protocol_name="",
        bundle=["a"],
        target_path=str(target),
    )
    plan = RefactorEngine(project_root=tmp_path).plan_protocol_extraction(request)
    assert plan.errors


def test_refactor_engine_requires_bundle_or_fields(tmp_path: Path) -> None:
    RefactorEngine, _, RefactorRequest, *_ = _load()
    target = tmp_path / "sample.py"
    target.write_text("def f(a):\n    return a\n")
    request = RefactorRequest(
        protocol_name="BundleProtocol",
        bundle=[],
        target_path=str(target),
    )
    plan = RefactorEngine(project_root=tmp_path).plan_protocol_extraction(request)
    assert plan.errors


def test_refactor_engine_invalid_type_hint_warns(tmp_path: Path) -> None:
    RefactorEngine, FieldSpec, RefactorRequest, *_ = _load()
    target = tmp_path / "sample.py"
    target.write_text("def f(a):\n    return a\n")
    request = RefactorRequest(
        protocol_name="BundleProtocol",
        bundle=["a"],
        fields=[FieldSpec(name="a", type_hint="list[")],
        target_path=str(target),
    )
    plan = RefactorEngine(project_root=tmp_path).plan_protocol_extraction(request)
    assert any("Failed to parse type hint" in warning for warning in plan.warnings)
    assert plan.edits


def test_module_name_strips_src_prefix(tmp_path: Path) -> None:
    _, _, _, _, _, _module_name, _ = _load()
    path = Path("src") / "pkg" / "mod.py"
    assert _module_name(path, None) == "pkg.mod"
    rooted = tmp_path / "pkg" / "mod.py"
    assert _module_name(rooted, tmp_path) == "pkg.mod"


def test_module_expr_to_str_and_import_index() -> None:
    _, _, _, _find_import_insert_index, _module_expr_to_str, _, _ = _load()
    module = cst.parse_module(
        textwrap.dedent(
            '''
            """Docstring."""
            import os

            def f():
                return 1
            '''
        ).strip()
        + "\n"
    )
    assert _find_import_insert_index(list(module.body)) == 2
    expr = cst.parse_expression("pkg.mod.util")
    assert _module_expr_to_str(expr) == "pkg.mod.util"
    assert _module_expr_to_str(None) is None


def test_has_typing_import_handles_attribute_module() -> None:
    _, _, _, _, _, _, _has_typing_import = _load()
    module = cst.parse_module("import foo.typing\n")
    assert _has_typing_import(list(module.body)) is False



from __future__ import annotations

from pathlib import Path
import sys


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.synthesis.model import NamingContext, SynthesisConfig
    from gabion.synthesis.protocols import Synthesizer

    return NamingContext, SynthesisConfig, Synthesizer


def test_synthesizer_filters_and_builds_specs() -> None:
    NamingContext, SynthesisConfig, Synthesizer = _load()
    synth = Synthesizer(config=SynthesisConfig(max_tier=2, min_bundle_size=2))
    bundle_tiers = {
        frozenset({"a", "b"}): 2,
        frozenset({"x"}): 2,
    }
    plan = synth.plan(bundle_tiers, field_types={"a": "int"})
    assert len(plan.protocols) == 1
    spec = plan.protocols[0]
    assert spec.name == "ABundle"
    assert spec.tier == 2
    assert {f.name for f in spec.fields} == {"a", "b"}
    assert next(f for f in spec.fields if f.name == "a").type_hint == "int"


def test_synthesizer_uses_existing_names() -> None:
    NamingContext, SynthesisConfig, Synthesizer = _load()
    synth = Synthesizer(
        config=SynthesisConfig(max_tier=2, min_bundle_size=1, allow_singletons=True)
    )
    bundle_tiers = {frozenset({"ctx"}): 2}
    context = NamingContext(existing_names={"CtxBundle"})
    plan = synth.plan(bundle_tiers, naming_context=context)
    assert plan.protocols[0].name == "CtxBundle2"


def test_synthesizer_warns_on_empty() -> None:
    NamingContext, SynthesisConfig, Synthesizer = _load()
    synth = Synthesizer(config=SynthesisConfig(max_tier=1, min_bundle_size=2))
    plan = synth.plan({})
    assert plan.protocols == []
    assert plan.warnings


def test_synthesizer_skips_empty_bundle() -> None:
    NamingContext, SynthesisConfig, Synthesizer = _load()
    synth = Synthesizer(config=SynthesisConfig(max_tier=1, min_bundle_size=1))
    plan = synth.plan({frozenset(): 1})
    assert plan.protocols == []



from __future__ import annotations

from pathlib import Path
import sys


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.synthesis.merge import _jaccard, merge_bundles

    return _jaccard, merge_bundles


def test_merge_bundles_combines_overlaps() -> None:
    _jaccard, merge_bundles = _load()
    bundles = [{"a", "b"}, {"a", "b", "c"}]
    merged = merge_bundles(bundles, min_overlap=0.66)
    assert merged == [{"a", "b", "c"}]
    assert _jaccard(set(), set()) == 1.0


def test_merge_bundles_keeps_distinct() -> None:
    _jaccard, merge_bundles = _load()
    bundles = [{"a", "b"}, {"c", "d"}]
    merged = merge_bundles(bundles, min_overlap=0.5)
    assert merged == [{"a", "b"}, {"c", "d"}]
    assert _jaccard(set(), {"x"}) == 0.0



#!/usr/bin/env python3
"""Sync SPPF checklist-linked issues from commit messages.

This is a local helper to avoid CI write scopes. It parses commit messages for
references like "GH-17" or "Closes #17" and comments/closes/labels issues
via the GitHub CLI.
"""
from __future__ import annotations

import argparse
import re
import subprocess
import sys
from dataclasses import dataclass


GH_REF_RE = re.compile(r"\bGH-(\d+)\b", re.IGNORECASE)
KEYWORD_REF_RE = re.compile(r"\b(?:Closes|Fixes|Resolves|Refs)\s+#(\d+)\b", re.IGNORECASE)


@dataclass(frozen=True)
class CommitInfo:
    sha: str
    subject: str
    body: str


def _run_git(args: list[str]) -> str:
    return subprocess.check_output(["git", *args], text=True).strip()


def _default_range() -> str:
    try:
        _run_git(["rev-parse", "origin/stage"])
        return "origin/stage..HEAD"
    except Exception:
        return "HEAD~20..HEAD"


def _collect_commits(rev_range: str) -> list[CommitInfo]:
    try:
        raw = subprocess.check_output(
            [
                "git",
                "log",
                "--format=%H%x1f%s%x1f%B%x1e",
                rev_range,
            ],
            text=True,
        )
    except subprocess.CalledProcessError as exc:
        raise SystemExit(f"git log failed for range {rev_range}: {exc}")

    commits: list[CommitInfo] = []
    for record in raw.split("\x1e"):
        if not record.strip():
            continue
        parts = record.split("\x1f")
        if len(parts) < 3:
            continue
        sha, subject, body = parts[0].strip(), parts[1].strip(), parts[2].strip()
        commits.append(CommitInfo(sha=sha, subject=subject, body=body))
    return commits


def _extract_issue_ids(text: str) -> set[str]:
    issues = set(match.group(1) for match in GH_REF_RE.finditer(text))
    issues.update(match.group(1) for match in KEYWORD_REF_RE.finditer(text))
    return issues


def _issue_ids_from_commits(commits: list[CommitInfo]) -> set[str]:
    issues: set[str] = set()
    for commit in commits:
        issues.update(_extract_issue_ids(commit.subject))
        issues.update(_extract_issue_ids(commit.body))
    return issues


def _build_comment(rev_range: str, commits: list[CommitInfo]) -> str:
    lines = [f"SPPF sync from `{rev_range}`:"]
    for commit in commits:
        lines.append(f"- {commit.sha[:8]} {commit.subject}")
    return "\n".join(lines)


def _run_gh(args: list[str], dry_run: bool) -> None:
    if dry_run:
        print("DRY RUN:", " ".join(["gh", *args]))
        return
    subprocess.run(["gh", *args], check=True)


def main() -> int:
    parser = argparse.ArgumentParser(description="Sync SPPF-linked issues from commit messages.")
    parser.add_argument(
        "--range",
        dest="rev_range",
        default=None,
        help="Git revision range (default: origin/stage..HEAD if available).")
    parser.add_argument(
        "--comment",
        action="store_true",
        help="Comment on each referenced issue with commit summary.")
    parser.add_argument(
        "--close",
        action="store_true",
        help="Close each referenced issue with a summary comment.")
    parser.add_argument(
        "--label",
        default=None,
        help="Apply a label to each referenced issue.")
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Print gh commands without executing.")
    args = parser.parse_args()

    rev_range = args.rev_range or _default_range()
    commits = _collect_commits(rev_range)
    if not commits:
        print("No commits in range; nothing to sync.")
        return 0

    issue_ids = sorted(_issue_ids_from_commits(commits))
    if not issue_ids:
        print("No issue references found in commit messages.")
        return 0

    comment = _build_comment(rev_range, commits)
    for issue_id in issue_ids:
        if args.close:
            _run_gh(["issue", "close", issue_id, "-c", comment], args.dry_run)
        elif args.comment:
            _run_gh(["issue", "comment", issue_id, "-b", comment], args.dry_run)
        if args.label:
            _run_gh(["issue", "edit", issue_id, "--add-label", args.label], args.dry_run)
    return 0


if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations

from pathlib import Path
import sys
import textwrap


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import analyze_paths, AuditConfig

    return analyze_paths, AuditConfig


def test_param_spans_capture_signature_positions(tmp_path: Path) -> None:
    analyze_paths, AuditConfig = _load()
    source = textwrap.dedent(
        """
        def f(a, b):
            return a + b
        """
    ).lstrip()
    file_path = tmp_path / "mod.py"
    file_path.write_text(source)
    config = AuditConfig(project_root=tmp_path)
    analysis = analyze_paths(
        [file_path],
        recursive=True,
        type_audit=False,
        type_audit_report=False,
        type_audit_max=0,
        include_constant_smells=False,
        include_unused_arg_smells=False,
        config=config,
    )
    spans = analysis.param_spans_by_path[file_path]["f"]
    assert "a" in spans
    assert "b" in spans
    line = source.splitlines()[0]
    a_col = line.index("a")
    b_col = line.index("b")
    a_span = spans["a"]
    b_span = spans["b"]
    assert a_span[0] == 0
    assert b_span[0] == 0
    assert a_span[1] == a_col
    assert b_span[1] == b_col
    assert a_span[3] > a_span[1]
    assert b_span[3] > b_span[1]



from __future__ import annotations

import io
import json

from gabion.lsp_client import LspClientError, _read_response, _read_rpc


def _rpc_message(payload: dict) -> bytes:
    body = json.dumps(payload).encode("utf-8")
    header = f"Content-Length: {len(body)}\r\n\r\n".encode("utf-8")
    return header + body


def test_read_rpc_invalid_length() -> None:
    stream = io.BytesIO(b"Content-Length: 0\r\n\r\n{}")
    try:
        _read_rpc(stream)
    except LspClientError as exc:
        assert "Content-Length" in str(exc)
    else:
        raise AssertionError("Expected LspClientError for invalid Content-Length")


def test_read_rpc_missing_content_length_header() -> None:
    stream = io.BytesIO(b"Foo: bar\r\n\r\n{}")
    try:
        _read_rpc(stream)
    except LspClientError as exc:
        assert "Content-Length" in str(exc)
    else:
        raise AssertionError("Expected LspClientError for missing Content-Length")


def test_read_rpc_stream_closed() -> None:
    stream = io.BytesIO(b"")
    try:
        _read_rpc(stream)
    except LspClientError as exc:
        assert "stream closed" in str(exc).lower()
    else:
        raise AssertionError("Expected LspClientError for closed stream")


def test_read_response_skips_unmatched_ids() -> None:
    msg1 = _rpc_message({"jsonrpc": "2.0", "id": 1, "result": {"ok": True}})
    msg2 = _rpc_message({"jsonrpc": "2.0", "id": 2, "result": {"answer": 42}})
    stream = io.BytesIO(msg1 + msg2)
    response = _read_response(stream, 2)
    assert response["id"] == 2
    assert response["result"]["answer"] == 42


def test_read_rpc_skips_non_content_length_headers() -> None:
    payload = {"jsonrpc": "2.0", "id": 1, "result": {}}
    body = json.dumps(payload).encode("utf-8")
    header = b"Foo: bar\r\nContent-Length: " + str(len(body)).encode("utf-8") + b"\r\n\r\n"
    stream = io.BytesIO(header + body)
    message = _read_rpc(stream)
    assert message["id"] == 1


def test_read_rpc_accepts_prefetched_body() -> None:
    payload = {"jsonrpc": "2.0", "id": 9, "result": {"ok": True}}
    body = json.dumps(payload).encode("utf-8")
    header = f"Content-Length: {len(body)}\r\n\r\n".encode("utf-8")

    class _Chunky:
        def __init__(self, data: bytes) -> None:
            self._data = data
            self._sent = False

        def read(self, _n: int) -> bytes:
            if self._sent:
                return b""
            self._sent = True
            return self._data

    stream = _Chunky(header + body)
    message = _read_rpc(stream)
    assert message["id"] == 9



from __future__ import annotations

import ast
from pathlib import Path
import sys
import types


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import (
        CallArgs,
        ParamUse,
        _callee_name,
        _call_context,
        _const_repr,
    )
    from gabion.analysis.visitors import ImportVisitor, ParentAnnotator, UseVisitor

    return (
        CallArgs,
        ParamUse,
        _callee_name,
        _call_context,
        _const_repr,
        ImportVisitor,
        ParentAnnotator,
        UseVisitor,
    )


def _make_use_visitor(
    code: str,
    params: list[str],
    *,
    strictness: str = "high",
    return_aliases: dict[str, tuple[list[str], list[str]]] | None = None,
):
    (
        CallArgs,
        ParamUse,
        _callee_name,
        _call_context,
        _const_repr,
        _,
        ParentAnnotator,
        UseVisitor,
    ) = _load()
    tree = ast.parse(code)
    annotator = ParentAnnotator()
    annotator.visit(tree)
    alias_to_param = {name: name for name in params}
    use_map = {
        name: ParamUse(direct_forward=set(), non_forward=False, current_aliases={name})
        for name in params
    }
    call_args: list[CallArgs] = []
    visitor = UseVisitor(
        parents=annotator.parents,
        use_map=use_map,
        call_args=call_args,
        alias_to_param=alias_to_param,
        is_test=False,
        strictness=strictness,
        const_repr=_const_repr,
        callee_name=_callee_name,
        call_args_factory=CallArgs,
        call_context=_call_context,
        return_aliases=return_aliases,
    )
    return tree, visitor, use_map, call_args


def test_import_visitor_relative_and_star() -> None:
    (
        _,
        _,
        _,
        _,
        _,
        ImportVisitor,
        _,
        _,
    ) = _load()
    table = types.SimpleNamespace(imports={}, star_imports={})
    visitor = ImportVisitor("pkg.mod", table)
    visitor.visit_ImportFrom(
        ast.ImportFrom(module=None, names=[ast.alias(name="x")], level=0)
    )
    visitor.visit_ImportFrom(
        ast.ImportFrom(module="too.deep", names=[ast.alias(name="y")], level=5)
    )
    visitor.visit_ImportFrom(
        ast.ImportFrom(module="utils", names=[ast.alias(name="*")], level=0)
    )
    visitor.visit_ImportFrom(
        ast.ImportFrom(module="helpers", names=[ast.alias(name="tool")], level=1)
    )
    assert table.star_imports["pkg.mod"] == {"utils"}
    assert table.imports[("pkg.mod", "tool")] == "pkg.helpers.tool"


def test_alias_from_call_branches() -> None:
    code = "def f(a, b):\n    return a\n"
    tree, visitor, _, _ = _make_use_visitor(
        code,
        ["a", "b"],
        return_aliases={
            "identity": (["x"], ["x"]),
            "noop": (["x"], []),
        },
    )
    call = ast.parse("noop(a)").body[0].value
    assert visitor._alias_from_call(call) is None
    call = ast.parse("identity(*args)").body[0].value
    assert visitor._alias_from_call(call) is None
    call = ast.parse("identity(a, b)").body[0].value
    assert visitor._alias_from_call(call) is None
    call = ast.parse("identity(x=1)").body[0].value
    assert visitor._alias_from_call(call) is None
    call = ast.parse("identity(a)").body[0].value
    assert visitor._alias_from_call(call) == ["a"]
    visitor.visit(tree)


def test_bind_sequence_and_return_alias_assignment() -> None:
    code = (
        "def f(a, b):\n"
        "    alias = identity(a)\n"
        "    extra: int = identity(a)\n"
        "    (x, y) = (a, b)\n"
    )
    tree, visitor, use_map, _ = _make_use_visitor(
        code,
        ["a", "b"],
        return_aliases={"identity": (["x"], ["x"])},
    )
    target = ast.parse("(x, y)").body[0].value
    rhs = ast.parse("(a,)").body[0].value
    assert visitor._bind_sequence(target, rhs) is False
    visitor.visit(tree)
    assert "alias" in visitor.alias_to_param
    assert use_map["a"].current_aliases


def test_star_args_strictness() -> None:
    code = "def f(a, b):\n    foo(*a)\n    foo(**b)\n"
    tree, visitor, use_map, _ = _make_use_visitor(code, ["a", "b"], strictness="high")
    visitor.visit(tree)
    assert use_map["a"].non_forward is True
    assert use_map["b"].non_forward is True
    tree, visitor, use_map, _ = _make_use_visitor(code, ["a", "b"], strictness="low")
    visitor.visit(tree)
    assert ("args[*]", "arg[*]") in use_map["a"].direct_forward
    assert ("kwargs[*]", "kw[*]") in use_map["b"].direct_forward


def test_attribute_and_subscript_aliases() -> None:
    code = (
        "def f(a, b):\n"
        "    alias = a\n"
        "    alias.attr = a\n"
        "    data = {}\n"
        "    data['k'] = b\n"
        "    foo(alias.attr)\n"
        "    foo(alias.other)\n"
        "    foo(alias['k'])\n"
        "    foo(alias[0])\n"
        "    foo(data['k'])\n"
        "    foo(alias.inner.attr)\n"
    )
    tree, visitor, use_map, _ = _make_use_visitor(code, ["a", "b"], strictness="low")
    visitor.visit(tree)
    assert any(slot == "arg[0]" for _, slot in use_map["a"].direct_forward)
    assert any(slot == "arg[0]" for _, slot in use_map["b"].direct_forward)
    assert use_map["a"].non_forward is True


def test_check_write_clears_attr_and_key_aliases() -> None:
    tree, visitor, use_map, _ = _make_use_visitor("def f(a):\n    pass\n", ["a"])
    visitor._attr_alias_to_param[("obj", "field")] = "a"
    visitor._key_alias_to_param[("data", "k")] = "a"
    visitor._check_write(ast.Name(id="obj", ctx=ast.Store()))
    visitor._check_write(ast.Name(id="data", ctx=ast.Store()))
    assert ("obj", "field") not in visitor._attr_alias_to_param
    assert ("data", "k") not in visitor._key_alias_to_param
    assert use_map["a"].non_forward is True


def test_bind_sequence_edge_cases() -> None:
    tree, visitor, _, _ = _make_use_visitor("def f(a, b):\n    pass\n", ["a", "b"])
    assert visitor._bind_sequence(
        ast.Name(id="x", ctx=ast.Store()),
        ast.Name(id="a", ctx=ast.Load()),
    ) is False
    target = ast.Tuple(
        elts=[ast.Name(id="x", ctx=ast.Store())],
        ctx=ast.Store(),
    )
    assert visitor._bind_sequence(target, ast.Name(id="a", ctx=ast.Load())) is False
    nested_target = ast.Tuple(
        elts=[
            ast.Tuple(
                elts=[ast.Name(id="x", ctx=ast.Store())],
                ctx=ast.Store(),
            )
        ],
        ctx=ast.Store(),
    )
    nested_rhs = ast.Tuple(
        elts=[
            ast.Tuple(
                elts=[ast.Name(id="a", ctx=ast.Load()), ast.Name(id="b", ctx=ast.Load())],
                ctx=ast.Load(),
            )
        ],
        ctx=ast.Load(),
    )
    assert visitor._bind_sequence(nested_target, nested_rhs) is True
    mismatch_target = ast.Tuple(
        elts=[ast.Name(id="x", ctx=ast.Store())],
        ctx=ast.Store(),
    )
    mismatch_rhs = ast.Tuple(
        elts=[ast.Constant(value=1)],
        ctx=ast.Load(),
    )
    assert visitor._bind_sequence(mismatch_target, mismatch_rhs) is True


def test_alias_from_call_keyword_branches() -> None:
    tree, visitor, _, _ = _make_use_visitor(
        "def f(a):\n    return a\n",
        ["a"],
        return_aliases={"identity": (["x"], ["x"])},
    )
    call = ast.parse("identity(**kwargs)").body[0].value
    assert visitor._alias_from_call(call) is None
    call = ast.parse("identity(extra=a)").body[0].value
    assert visitor._alias_from_call(call) is None
    call = ast.parse("identity(x=1)").body[0].value
    assert visitor._alias_from_call(call) is None
    visitor.visit(tree)


def test_alias_from_call_positional_and_kw_aliases() -> None:
    tree, visitor, _, _ = _make_use_visitor(
        "def f(a):\n    return a\n",
        ["a"],
        return_aliases={"identity": (["x"], ["x"])},
    )
    call = ast.parse("identity(1)").body[0].value
    assert visitor._alias_from_call(call) is None
    call = ast.parse("identity(x=a)").body[0].value
    assert visitor._alias_from_call(call) == ["a"]
    visitor.visit(tree)


def test_bind_return_alias_rejects_invalid_targets() -> None:
    tree, visitor, _, _ = _make_use_visitor("def f(a):\n    return a\n", ["a"])
    targets = [
        ast.Name(id="x", ctx=ast.Store()),
        ast.Name(id="y", ctx=ast.Store()),
    ]
    assert visitor._bind_return_alias(targets, ["a"]) is False
    assert visitor._bind_return_alias([ast.Name(id="x", ctx=ast.Store())], ["a", "b"]) is False
    tuple_target = ast.Tuple(
        elts=[ast.Name(id="x", ctx=ast.Store()), ast.Constant(value=1)],
        ctx=ast.Store(),
    )
    assert visitor._bind_return_alias([tuple_target], ["a", "b"]) is False
    assert visitor._bind_return_alias([ast.Constant(value=1)], ["a"]) is False
    visitor.visit(tree)


def test_annassign_edges() -> None:
    tree, visitor, use_map, _ = _make_use_visitor(
        "def f(a):\n    x: int = a\n",
        ["a"],
        return_aliases={"identity": (["x"], ["x"])},
    )
    visitor.visit(tree)
    assert "x" in visitor.alias_to_param
    assert use_map["a"].current_aliases
    visitor.visit_AnnAssign(
        ast.AnnAssign(
            target=ast.Name(id="y", ctx=ast.Store()),
            annotation=ast.Name(id="int", ctx=ast.Load()),
            value=None,
            simple=1,
        )
    )
    visitor.visit_AnnAssign(
        ast.AnnAssign(
            target=ast.Attribute(
                value=ast.Name(id="obj", ctx=ast.Load()),
                attr="attr",
                ctx=ast.Store(),
            ),
            annotation=ast.Name(id="int", ctx=ast.Load()),
            value=ast.Constant(value=1),
            simple=0,
        )
    )


def test_visit_name_attribute_subscript_edges() -> None:
    tree, visitor, use_map, _ = _make_use_visitor("def f(a):\n    pass\n", ["a"])
    visitor.visit_Name(ast.Name(id="a", ctx=ast.Store()))
    assert visitor._root_name(ast.Constant(value=1)) is None
    visitor._attr_alias_to_param[("obj", "field")] = "a"
    visitor.alias_to_param["obj"] = "a"
    visitor._suspend_non_forward.add("a")
    visitor.visit_Attribute(
        ast.Attribute(
            value=ast.Name(id="obj", ctx=ast.Load()),
            attr="field",
            ctx=ast.Load(),
        )
    )
    visitor._suspend_non_forward.clear()
    visitor.visit_Attribute(
        ast.Attribute(
            value=ast.Attribute(
                value=ast.Name(id="a", ctx=ast.Load()),
                attr="b",
                ctx=ast.Load(),
            ),
            attr="c",
            ctx=ast.Load(),
        )
    )
    assert use_map["a"].non_forward is True
    visitor.visit_Attribute(
        ast.Attribute(
            value=ast.Name(id="a", ctx=ast.Load()),
            attr="b",
            ctx=ast.Store(),
        )
    )
    visitor.visit_Subscript(
        ast.Subscript(
            value=ast.Name(id="a", ctx=ast.Load()),
            slice=ast.Name(id="k", ctx=ast.Load()),
            ctx=ast.Load(),
        )
    )
    visitor.visit_Subscript(
        ast.Subscript(
            value=ast.Name(id="a", ctx=ast.Load()),
            slice=ast.Constant(value="missing"),
            ctx=ast.Load(),
        )
    )
    visitor._key_alias_to_param[("a", "k")] = "a"
    visitor._suspend_non_forward.add("a")
    visitor.visit_Subscript(
        ast.Subscript(
            value=ast.Name(id="a", ctx=ast.Load()),
            slice=ast.Constant(value="k"),
            ctx=ast.Load(),
        )
    )
    visitor._suspend_non_forward.clear()
    visitor.visit_Subscript(
        ast.Subscript(
            value=ast.Name(id="a", ctx=ast.Load()),
            slice=ast.Constant(value="k"),
            ctx=ast.Store(),
        )
    )


def test_collect_alias_sources_default() -> None:
    tree, visitor, _, _ = _make_use_visitor("def f(a):\n    pass\n", ["a"])
    assert visitor._collect_alias_sources(ast.Constant(value=1)) == set()
    visitor.visit(tree)


def test_attr_and_subscript_slot_fallbacks_with_aliases() -> None:
    code = (
        "def f(a, data, obj):\n"
        "    obj.attr = a\n"
        "    data['k'] = a\n"
        "    foo(**obj.attr)\n"
        "    bar(**data['k'])\n"
    )
    tree, visitor, use_map, _ = _make_use_visitor(code, ["a", "data", "obj"])
    visitor.visit(tree)
    assert ("foo", "arg[?]") in use_map["a"].direct_forward
    assert ("bar", "arg[?]") in use_map["a"].direct_forward


def test_subscript_non_name_root_sets_non_forward() -> None:
    code = "def f(a):\n    foo(a.b['k'])\n"
    tree, visitor, use_map, _ = _make_use_visitor(code, ["a"])
    visitor.visit(tree)
    assert use_map["a"].non_forward is True


def test_subscript_and_attribute_slot_fallbacks() -> None:
    code = (
        "def f(a, data, obj):\n"
        "    foo(**data['k'])\n"
        "    bar(**obj.attr)\n"
    )
    tree, visitor, _, _ = _make_use_visitor(code, ["a", "data", "obj"])
    visitor.visit(tree)



#!/usr/bin/env python3
import argparse
import json
import os
import re
import sys
from dataclasses import dataclass
from pathlib import Path
from urllib.error import HTTPError, URLError
from urllib.request import Request, urlopen

try:
    import yaml
except ImportError:  # pragma: no cover - handled as a hard error at runtime.
    yaml = None


REPO_ROOT = Path(__file__).resolve().parents[1]
WORKFLOW_DIR = REPO_ROOT / ".github" / "workflows"

ALLOWED_ACTIONS_FILE = REPO_ROOT / "docs" / "allowed_actions.txt"
REQUIRED_RUNNER_LABELS = {"self-hosted", "gpu", "local"}
TRUSTED_BRANCHES = {"main", "stage", "next", "release"}
CONTENT_WRITE_WORKFLOWS = {
    "auto-test-tag.yml",
    "release-tag.yml",
    "mirror-next.yml",
    "promote-release.yml",
}

_SHA_RE = re.compile(r"^[0-9a-fA-F]{40}$")


@dataclass(frozen=True)
class JobContext:
    job_name: str
    path: Path


def _fail(errors):
    for err in errors:
        print(f"policy-check: {err}", file=sys.stderr)
    raise SystemExit(2)


def _load_allowed_actions(path: Path) -> set[str]:
    if not path.exists():
        return set()
    try:
        raw = path.read_text()
    except OSError:
        return set()
    allowed: set[str] = set()
    for line in raw.splitlines():
        line = line.strip()
        if not line or line.startswith("#"):
            continue
        allowed.add(line)
    return allowed


def _yaml_loader():
    if yaml is None:
        return None

    class Loader(yaml.SafeLoader):
        pass

    # Treat "on"/"off"/"yes"/"no" as strings, not booleans (YAML 1.1 quirk).
    for key, values in list(Loader.yaml_implicit_resolvers.items()):
        Loader.yaml_implicit_resolvers[key] = [
            (tag, regexp) for tag, regexp in values if tag != "tag:yaml.org,2002:bool"
        ]
    return Loader


def _load_yaml(path):
    if yaml is None:
        _fail(["PyYAML is required for policy checks (pip install pyyaml)."])
    with path.open("r", encoding="utf-8") as handle:
        loader = _yaml_loader()
        return yaml.load(handle, Loader=loader) or {}


def _normalize_if(expr):
    if not expr:
        return ""
    return re.sub(r"\s+", "", str(expr))


def _runs_on_labels(runs_on):
    if runs_on is None:
        return []
    if isinstance(runs_on, str):
        return [runs_on]
    if isinstance(runs_on, list):
        return [str(item) for item in runs_on]
    if isinstance(runs_on, dict):
        labels = runs_on.get("labels")
        if isinstance(labels, list):
            return [str(item) for item in labels]
        if isinstance(labels, str):
            return [labels]
    return []


def _is_self_hosted(runs_on):
    labels = _runs_on_labels(runs_on)
    return "self-hosted" in labels


def _event_names(on_block):
    if isinstance(on_block, str):
        return {on_block}
    if isinstance(on_block, list):
        return {str(item) for item in on_block}
    if isinstance(on_block, dict):
        return {str(key) for key in on_block.keys()}
    return set()


def _has_tag_push(on_block) -> bool:
    if not isinstance(on_block, dict):
        return False
    push_block = on_block.get("push")
    if push_block is None:
        return False
    if isinstance(push_block, dict):
        tags = push_block.get("tags")
        if isinstance(tags, list):
            return len(tags) > 0
        if isinstance(tags, str):
            return True
    return False


def _check_permissions(
    doc,
    path,
    errors,
    *,
    allow_pr_write=False,
    allow_id_token=False,
    allow_contents_write=False,
):
    # dataflow-bundle: doc, errors
    permissions = doc.get("permissions")
    if permissions is None:
        errors.append(f"{path}: missing top-level permissions")
        return
    if isinstance(permissions, str):
        errors.append(f"{path}: permissions must be a mapping, not {permissions!r}")
        return
    contents = permissions.get("contents")
    if allow_contents_write:
        if contents != "write":
            errors.append(f"{path}: permissions.contents must be 'write'")
    elif contents != "read":
        errors.append(f"{path}: permissions.contents must be 'read'")
    for key, value in permissions.items():
        if key == "contents":
            continue
        if value in ("read", "none"):
            continue
        if allow_pr_write and key == "pull-requests" and value == "write":
            continue
        if allow_id_token and key == "id-token" and value == "write":
            continue
        if allow_contents_write and key == "contents" and value == "write":
            continue
        errors.append(f"{path}: permissions.{key} must be 'read' or 'none'")


def _check_job_permissions(
    job,
    job_ctx: JobContext,
    errors,
    *,
    allow_pr_write=False,
    allow_id_token=False,
    allow_contents_write=False,
):
    # dataflow-bundle: errors, job, job_ctx
    permissions = job.get("permissions")
    if permissions is None:
        return
    if isinstance(permissions, str):
        errors.append(
            f"{job_ctx.path}:{job_ctx.job_name}: job permissions must be a mapping"
        )
        return
    contents = permissions.get("contents")
    if allow_contents_write:
        if contents != "write":
            errors.append(
                f"{job_ctx.path}:{job_ctx.job_name}: permissions.contents must be 'write'"
            )
    elif contents != "read":
        errors.append(
            f"{job_ctx.path}:{job_ctx.job_name}: permissions.contents must be 'read'"
        )
    is_self_hosted = _is_self_hosted(job.get("runs-on"))
    for key, value in permissions.items():
        if key == "contents":
            continue
        if value in ("read", "none"):
            continue
        if (
            allow_pr_write
            and not is_self_hosted
            and key == "pull-requests"
            and value == "write"
        ):
            continue
        if allow_id_token and not is_self_hosted and key == "id-token" and value == "write":
            continue
        if allow_contents_write and key == "contents" and value == "write":
            continue
        errors.append(
            f"{job_ctx.path}:{job_ctx.job_name}: permissions.{key} must be 'read' or 'none'"
        )


def _check_release_tag_workflow(doc, path, errors):
    events = _event_names(doc.get("on"))
    if events != {"workflow_dispatch"}:
        errors.append(f"{path}: release tag workflow must use workflow_dispatch only")
    jobs = doc.get("jobs", {})
    if not isinstance(jobs, dict):
        return
    for name, job in jobs.items():
        if _is_self_hosted(job.get("runs-on")):
            errors.append(f"{path}:{name}: release tag workflow must not use self-hosted")
        cond = _normalize_if(job.get("if"))
        if (
            "github.ref=='refs/heads/release'" not in cond
            and "github.ref=='refs/heads/next'" not in cond
        ):
            errors.append(
                f"{path}:{name}: release tag workflow must guard on refs/heads/release or refs/heads/next"
            )
        if "github.actor==github.repository_owner" not in cond:
            errors.append(
                f"{path}:{name}: release tag workflow must guard on repository owner"
            )


def _check_auto_test_tag_workflow(doc, path, errors):
    events = _event_names(doc.get("on"))
    if events != {"workflow_run"}:
        errors.append(f"{path}: auto test tag workflow must use workflow_run only")
    workflow_run = None
    if isinstance(doc.get("on"), dict):
        workflow_run = doc.get("on").get("workflow_run")
    workflows = None
    if isinstance(workflow_run, dict):
        workflows = workflow_run.get("workflows")
    if not workflows:
        errors.append(f"{path}: auto test tag workflow must specify workflows")
    jobs = doc.get("jobs", {})
    if not isinstance(jobs, dict):
        return
    for name, job in jobs.items():
        if _is_self_hosted(job.get("runs-on")):
            errors.append(f"{path}:{name}: auto test tag workflow must not use self-hosted")
        cond = _normalize_if(job.get("if"))
        if "github.event.workflow_run.conclusion=='success'" not in cond:
            errors.append(f"{path}:{name}: auto test tag workflow must guard on success")
        if "github.event.workflow_run.head_branch=='main'" not in cond:
            errors.append(f"{path}:{name}: auto test tag workflow must guard on main")
        if (
            "github.event.workflow_run.actor.login==github.repository_owner" not in cond
            and "github.event.workflow_run.actor.login=='github-actions[bot]'" not in cond
        ):
            errors.append(f"{path}:{name}: auto test tag workflow must guard on actor")


def _check_mirror_next_workflow(doc, path, errors):
    events = _event_names(doc.get("on"))
    if events != {"push"}:
        errors.append(f"{path}: mirror workflow must use push only")
    push_block = None
    if isinstance(doc.get("on"), dict):
        push_block = doc.get("on").get("push")
    branches = None
    if isinstance(push_block, dict):
        branches = push_block.get("branches")
    if not branches or ("main" not in branches):
        errors.append(f"{path}: mirror workflow must target main branch pushes")
    jobs = doc.get("jobs", {})
    if not isinstance(jobs, dict):
        return
    for name, job in jobs.items():
        if _is_self_hosted(job.get("runs-on")):
            errors.append(f"{path}:{name}: mirror workflow must not use self-hosted")
        cond = _normalize_if(job.get("if"))
        if "github.ref=='refs/heads/main'" not in cond:
            errors.append(f"{path}:{name}: mirror workflow must guard on main")
        if "github.actor==github.repository_owner" not in cond:
            errors.append(
                f"{path}:{name}: mirror workflow must guard on repository owner"
            )


def _check_promote_release_workflow(doc, path, errors):
    events = _event_names(doc.get("on"))
    if events != {"workflow_run"}:
        errors.append(f"{path}: promote workflow must use workflow_run only")
    workflow_run = None
    if isinstance(doc.get("on"), dict):
        workflow_run = doc.get("on").get("workflow_run")
    workflows = None
    if isinstance(workflow_run, dict):
        workflows = workflow_run.get("workflows")
    if not workflows:
        errors.append(f"{path}: promote workflow must specify workflows")
    jobs = doc.get("jobs", {})
    if not isinstance(jobs, dict):
        return
    for name, job in jobs.items():
        if _is_self_hosted(job.get("runs-on")):
            errors.append(f"{path}:{name}: promote workflow must not use self-hosted")
        cond = _normalize_if(job.get("if"))
        if "github.event.workflow_run.conclusion=='success'" not in cond:
            errors.append(f"{path}:{name}: promote workflow must guard on success")
        steps = job.get("steps", [])
        has_tag_artifact = False
        if isinstance(steps, list):
            for step in steps:
                uses = step.get("uses", "")
                if uses.startswith("actions/download-artifact@"):
                    with_block = step.get("with", {}) or {}
                    if with_block.get("name") == "release-testpypi-tag":
                        has_tag_artifact = True
                        break
                run = step.get("run", "")
                if "release-testpypi/tag.txt" in run:
                    has_tag_artifact = True
                    break
        if (
            "startswith(github.event.workflow_run.head_branch,'test-v')" not in cond
            and "startsWith(github.event.workflow_run.head_branch,'test-v')" not in cond
            and not has_tag_artifact
        ):
            errors.append(f"{path}:{name}: promote workflow must guard on test-v tags")
        if (
            "github.event.workflow_run.actor.login==github.repository_owner" not in cond
            and "github.event.workflow_run.actor.login=='github-actions[bot]'" not in cond
        ):
            errors.append(
                f"{path}:{name}: promote workflow must guard on repository owner or github-actions[bot]"
            )


def _step_run_contains_tokens(steps, tokens: set[str]) -> bool:
    for step in steps:
        if not isinstance(step, dict):
            continue
        run = step.get("run")
        if not isinstance(run, str):
            continue
        if all(token in run for token in tokens):
            return True
    return False


def _step_run_contains_any(steps, tokens: set[str]) -> bool:
    for step in steps:
        if not isinstance(step, dict):
            continue
        run = step.get("run")
        if not isinstance(run, str):
            continue
        if any(token in run for token in tokens):
            return True
    return False


def _check_release_testpypi_workflow(doc, path, errors):
    jobs = doc.get("jobs", {})
    if not isinstance(jobs, dict):
        return
    required_tokens = {
        "refs/tags/$TAG",
        "origin/main",
        "origin/next",
        "TAG_SHA",
    }
    required_scripts = {
        "scripts/release_verify_test_tag.py",
    }
    for name, job in jobs.items():
        steps = job.get("steps", [])
        if not (
            _step_run_contains_tokens(steps, required_tokens)
            or _step_run_contains_any(steps, required_scripts)
        ):
            errors.append(
                f"{path}:{name}: release-testpypi workflow must verify tag equals main/next"
            )


def _check_release_pypi_workflow(doc, path, errors):
    jobs = doc.get("jobs", {})
    if not isinstance(jobs, dict):
        return
    required_tokens = {
        "origin/main",
        "origin/next",
        "origin/release",
        "TAG_SHA",
    }
    required_scripts = {
        "scripts/release_verify_pypi_tag.py",
    }
    for name, job in jobs.items():
        steps = job.get("steps", [])
        if not (
            _step_run_contains_tokens(steps, required_tokens)
            or _step_run_contains_any(steps, required_scripts)
        ):
            errors.append(
                f"{path}:{name}: release-pypi workflow must verify tag equals main/next/release"
            )


def _check_actions(job, job_ctx: JobContext, errors, allowed_actions: set[str]):
    # dataflow-bundle: errors, job, job_ctx
    steps = job.get("steps", [])
    for idx, step in enumerate(steps):
        uses = step.get("uses")
        if not uses:
            continue
        if uses.startswith("./") or uses.startswith("docker://"):
            continue
        if "@" not in uses:
            errors.append(
                f"{job_ctx.path}:{job_ctx.job_name}: step {idx} uses unpinned action {uses!r}"
            )
            continue
        action_ref, ref = uses.split("@", 1)
        action_parts = action_ref.split("/")
        if len(action_parts) < 2:
            errors.append(
                f"{job_ctx.path}:{job_ctx.job_name}: step {idx} invalid action {uses!r}"
            )
            continue
        action_name = "/".join(action_parts[:2])
        if action_name not in allowed_actions:
            errors.append(
                f"{job_ctx.path}:{job_ctx.job_name}: step {idx} action not allow-listed ({action_name})"
            )
        if not _SHA_RE.match(ref):
            errors.append(
                f"{job_ctx.path}:{job_ctx.job_name}: step {idx} action not pinned to full SHA ({uses})"
            )


def _check_self_hosted_constraints(doc, path, errors):
    # dataflow-bundle: doc, errors
    jobs = doc.get("jobs", {})
    if not isinstance(jobs, dict):
        return
    self_hosted_jobs = []
    for name, job in jobs.items():
        if _is_self_hosted(job.get("runs-on")):
            self_hosted_jobs.append((name, job))

    if not self_hosted_jobs:
        return

    on_block = doc.get("on")
    events = _event_names(on_block)
    if events != {"push"}:
        errors.append(f"{path}: self-hosted workflow must trigger only on push")
    push_block = None
    if isinstance(on_block, dict):
        push_block = on_block.get("push")
    branches = None
    if isinstance(push_block, dict):
        branches = push_block.get("branches")
    if not branches:
        errors.append(f"{path}: push trigger must restrict branches")
    else:
        branch_set = {str(item) for item in branches}
        if not branch_set.issubset(TRUSTED_BRANCHES):
            errors.append(
                f"{path}: push branches must be subset of {sorted(TRUSTED_BRANCHES)}"
            )

    for name, job in self_hosted_jobs:
        runs_on = job.get("runs-on")
        labels = set(_runs_on_labels(runs_on))
        if not labels:
            errors.append(f"{path}:{name}: runs-on must be explicit labels")
        if not REQUIRED_RUNNER_LABELS.issubset(labels):
            errors.append(
                f"{path}:{name}: runs-on must include {sorted(REQUIRED_RUNNER_LABELS)}"
            )
        cond = _normalize_if(job.get("if"))
        if "github.actor==github.repository_owner" not in cond:
            errors.append(
                f"{path}:{name}: missing actor guard (github.actor == github.repository_owner)"
            )


def check_workflows():
    allowed_actions = _load_allowed_actions(ALLOWED_ACTIONS_FILE)
    if not allowed_actions:
        _fail([f"allowed actions list is empty or missing: {ALLOWED_ACTIONS_FILE}"])
    errors = []
    for path in sorted(WORKFLOW_DIR.glob("*.yml")):
        doc = _load_yaml(path)
        jobs = doc.get("jobs", {})
        has_self_hosted = False
        if isinstance(jobs, dict):
            for name, job in jobs.items():
                if _is_self_hosted(job.get("runs-on")):
                    has_self_hosted = True
        events = _event_names(doc.get("on"))
        allow_id_token = _has_tag_push(doc.get("on")) and (not has_self_hosted)
        allow_pr_write = (("pull_request" in events) or ("pull_request_target" in events)) and (
            not has_self_hosted
        )
        allow_contents_write = path.name in CONTENT_WRITE_WORKFLOWS
        if allow_contents_write:
            if path.name == "release-tag.yml":
                _check_release_tag_workflow(doc, path, errors)
            if path.name == "auto-test-tag.yml":
                _check_auto_test_tag_workflow(doc, path, errors)
            if path.name == "mirror-next.yml":
                _check_mirror_next_workflow(doc, path, errors)
            if path.name == "promote-release.yml":
                _check_promote_release_workflow(doc, path, errors)
        if path.name == "release-testpypi.yml":
            _check_release_testpypi_workflow(doc, path, errors)
        if path.name == "release-pypi.yml":
            _check_release_pypi_workflow(doc, path, errors)
        _check_permissions(
            doc,
            path,
            errors,
            allow_pr_write=allow_pr_write,
            allow_id_token=allow_id_token,
            allow_contents_write=allow_contents_write,
        )
        _check_self_hosted_constraints(doc, path, errors)
        if isinstance(jobs, dict):
            for name, job in jobs.items():
                job_ctx = JobContext(job_name=str(name), path=path)
                _check_job_permissions(
                    job,
                    job_ctx,
                    errors,
                    allow_pr_write=allow_pr_write,
                    allow_id_token=allow_id_token,
                    allow_contents_write=allow_contents_write,
                )
                _check_actions(job, job_ctx, errors, allowed_actions)
    if errors:
        _fail(errors)


def _api_json(url, token):
    headers = {
        "Accept": "application/vnd.github+json",
        "User-Agent": "policy-check",
        "Authorization": f"Bearer {token}",
    }
    request = Request(url, headers=headers)
    with urlopen(request) as response:
        return json.loads(response.read().decode("utf-8"))


def check_posture():
    policy_token = os.environ.get("POLICY_GITHUB_TOKEN")
    token = policy_token or os.environ.get("GITHUB_TOKEN")
    using_policy_token = policy_token is not None
    repo = os.environ.get("GITHUB_REPOSITORY")
    if not token:
        _fail(["missing GITHUB_TOKEN or POLICY_GITHUB_TOKEN for posture check"])
    if policy_token is not None and len(policy_token) < 20:
        _fail(
            [
                "POLICY_GITHUB_TOKEN appears too short; reset the repo secret "
                "with scripts/reset_policy_secret.sh"
            ]
        )
    if not repo:
        _fail(["missing GITHUB_REPOSITORY for posture check"])

    base = f"https://api.github.com/repos/{repo}"
    errors = []
    try:
        perms = _api_json(f"{base}/actions/permissions", token)
    except (HTTPError, URLError) as exc:
        hint = ""
        if isinstance(exc, HTTPError):
            if exc.code == 401:
                hint = " (POLICY_GITHUB_TOKEN is invalid or missing required auth)"
            elif exc.code == 403 and not using_policy_token:
                hint = " (set POLICY_GITHUB_TOKEN with admin read access)"
        _fail([f"actions permissions query failed: {exc}{hint}"])
    allowed = perms.get("allowed_actions")
    if allowed != "selected":
        errors.append("allowed_actions must be 'selected'")

    try:
        workflow_perms = _api_json(f"{base}/actions/permissions/workflow", token)
    except (HTTPError, URLError) as exc:
        hint = ""
        if isinstance(exc, HTTPError):
            if exc.code == 401:
                hint = " (POLICY_GITHUB_TOKEN is invalid or missing required auth)"
            elif exc.code == 403 and not using_policy_token:
                hint = " (set POLICY_GITHUB_TOKEN with admin read access)"
        _fail([f"workflow permissions query failed: {exc}{hint}"])
    default_perms = workflow_perms.get("default_workflow_permissions")
    if default_perms != "read":
        errors.append("default_workflow_permissions must be 'read'")
    if workflow_perms.get("can_approve_pull_request_reviews"):
        errors.append("can_approve_pull_request_reviews must be false")

    try:
        selected = _api_json(
            f"{base}/actions/permissions/selected-actions", token
        )
    except (HTTPError, URLError) as exc:
        hint = ""
        if isinstance(exc, HTTPError):
            if exc.code == 401:
                hint = " (POLICY_GITHUB_TOKEN is invalid or missing required auth)"
            elif exc.code == 403 and not using_policy_token:
                hint = " (set POLICY_GITHUB_TOKEN with admin read access)"
        _fail([f"selected actions query failed: {exc}{hint}"])

    if selected.get("github_owned_allowed") or selected.get("verified_allowed"):
        errors.append("only allow-listed actions are permitted (disable broad allowances)")

    patterns = selected.get("patterns_allowed") or []
    normalized = []
    invalid_patterns = []
    for pattern in patterns:
        if "@" in pattern:
            name, ref = pattern.split("@", 1)
            if ref != "*":
                invalid_patterns.append(pattern)
                continue
            normalized.append(name)
        else:
            normalized.append(pattern)
    if invalid_patterns:
        errors.append(
            f"invalid action patterns (must end with @* or be bare): {sorted(invalid_patterns)}"
        )
    allowed_actions = _load_allowed_actions(ALLOWED_ACTIONS_FILE)
    if set(normalized) != set(allowed_actions):
        errors.append(
            f"allowed action patterns must match {sorted(allowed_actions)}"
        )

    if errors:
        _fail(errors)


def main():
    parser = argparse.ArgumentParser(description="POLICY_SEED guardrails")
    parser.add_argument("--workflows", action="store_true", help="lint workflows")
    parser.add_argument("--posture", action="store_true", help="check GitHub posture")
    args = parser.parse_args()

    if not args.workflows and not args.posture:
        args.workflows = True

    if args.workflows:
        check_workflows()
    if args.posture:
        check_posture()
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Set


@dataclass(frozen=True)
class FieldSpec:
    name: str
    type_hint: str | None = None
    source_params: Set[str] = field(default_factory=set)


@dataclass(frozen=True)
class ProtocolSpec:
    name: str
    fields: List[FieldSpec]
    bundle: Set[str]
    tier: int
    rationale: str | None = None


@dataclass(frozen=True)
class SynthesisPlan:
    protocols: List[ProtocolSpec] = field(default_factory=list)
    warnings: List[str] = field(default_factory=list)
    errors: List[str] = field(default_factory=list)


@dataclass(frozen=True)
class NamingContext:
    existing_names: Set[str] = field(default_factory=set)
    frequency: Dict[str, int] = field(default_factory=dict)
    fallback_prefix: str = "Bundle"


@dataclass(frozen=True)
class SynthesisConfig:
    max_tier: int = 2
    min_bundle_size: int = 2
    allow_singletons: bool = False
    merge_overlap_threshold: float = 0.75



from __future__ import annotations

from pathlib import Path
import sys
import textwrap


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import AuditConfig, analyze_paths, build_synthesis_plan

    return AuditConfig, analyze_paths, build_synthesis_plan


def test_knob_merge_by_constant_params(tmp_path: Path) -> None:
    AuditConfig, analyze_paths, build_synthesis_plan = _load()
    source = textwrap.dedent(
        """
        def sink(a, b, mode):
            return a, b, mode

        def f(a, b, mode):
            return sink(a, b, mode)

        def g(a, b):
            return f(a, b, "fast")
        """
    )
    file_path = tmp_path / "mod.py"
    file_path.write_text(source.strip() + "\n")
    config = AuditConfig(project_root=tmp_path)
    groups_by_path = {
        file_path: {
            "f": [set(["a", "b", "mode"])],
            "g": [set(["a", "b"])],
        }
    }
    plan = build_synthesis_plan(
        groups_by_path,
        project_root=tmp_path,
        max_tier=3,
        merge_overlap_threshold=0.9,
        config=config,
    )
    bundles = [set(protocol["bundle"]) for protocol in plan.get("protocols", [])]
    assert {"a", "b", "mode"} in bundles
    assert {"a", "b"} not in bundles



from __future__ import annotations

from pathlib import Path
import sys
import textwrap


def _write(tmp_path: Path, rel: str, content: str) -> Path:
    path = tmp_path / rel
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(textwrap.dedent(content))
    return path


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import analyze_unused_arg_flow_repo

    return analyze_unused_arg_flow_repo


def test_star_import_resolution_disambiguates(tmp_path: Path) -> None:
    analyze_unused_arg_flow_repo = _load()
    _write(
        tmp_path,
        "a.py",
        """
        def helper(a, b):
            return b
        """,
    )
    _write(
        tmp_path,
        "c.py",
        """
        def helper(a, b):
            return a + b
        """,
    )
    b_path = _write(
        tmp_path,
        "b.py",
        """
        from a import *

        def caller(x, y):
            return helper(x, y)
        """,
    )
    smells = analyze_unused_arg_flow_repo(
        [tmp_path / "a.py", tmp_path / "b.py", tmp_path / "c.py"],
        project_root=tmp_path,
        ignore_params=set(),
        strictness="high",
        external_filter=True,
    )
    assert any(
        "b.py:caller passes param x to unused a.py:helper.a" in smell for smell in smells
    )


def test_star_import_resolves_reexport(tmp_path: Path) -> None:
    analyze_unused_arg_flow_repo = _load()
    _write(
        tmp_path,
        "a.py",
        """
        def helper(a, b):
            return b
        """,
    )
    _write(
        tmp_path,
        "b.py",
        """
        from a import helper
        __all__ = ["helper"]
        """,
    )
    _write(
        tmp_path,
        "c.py",
        """
        from b import *

        def caller(x, y):
            return helper(x, y)
        """,
    )
    smells = analyze_unused_arg_flow_repo(
        [tmp_path / "a.py", tmp_path / "b.py", tmp_path / "c.py"],
        project_root=tmp_path,
        ignore_params=set(),
        strictness="high",
        external_filter=True,
    )
    assert any(
        "c.py:caller passes param x to unused a.py:helper.a" in smell for smell in smells
    )



from __future__ import annotations

import argparse
from pathlib import Path
import re


def _normalize_test_version(tag: str) -> str:
    if not tag.startswith("test-v"):
        raise SystemExit(f"Unexpected tag: {tag}")
    version = tag[len("test-v") :]
    if not version:
        raise SystemExit("Missing version in tag")
    if "+" in version:
        base, local = version.split("+", 1)
        digits = "".join(ch for ch in local if ch.isdigit())
        if not digits:
            raise SystemExit("Missing timestamp digits in test tag")
        version = f"{base}.dev{digits}"
    return version


def _update_pyproject(version: str, path: Path) -> None:
    text = path.read_text(encoding="utf-8")
    updated, count = re.subn(
        r'(?m)^version = ".*"$', f'version = "{version}"', text, count=1
    )
    if count != 1:
        raise SystemExit("Failed to update project.version in pyproject.toml")
    path.write_text(updated, encoding="utf-8")


def _update_init(version: str, path: Path) -> None:
    text = path.read_text(encoding="utf-8").replace("\r\n", "\n")
    pattern = r'(?m)^__version__\s*=\s*([\'\"]).*?\1\s*$'
    updated, count = re.subn(pattern, f'__version__ = "{version}"', text, count=1)
    if count != 1:
        print(
            "Warning: __version__ update skipped "
            f"(matches={count})"
        )
        return
    path.write_text(updated, encoding="utf-8")


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("--tag", required=True)
    args = parser.parse_args()

    version = _normalize_test_version(args.tag)
    _update_pyproject(version, Path("pyproject.toml"))
    _update_init(version, Path("src/gabion/__init__.py"))
    print(f"Set project.version to {version}")


if __name__ == "__main__":
    main()



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Iterable, List, Mapping, Set

from gabion.synthesis.model import (
    FieldSpec,
    NamingContext,
    ProtocolSpec,
    SynthesisConfig,
    SynthesisPlan,
)
from gabion.synthesis.naming import suggest_name


@dataclass
class Synthesizer:
    config: SynthesisConfig = field(default_factory=SynthesisConfig)

    def plan(
        self,
        bundle_tiers: Mapping[frozenset[str], int],
        field_types: Mapping[str, str] | None = None,
        naming_context: NamingContext | None = None,
    ) -> SynthesisPlan:
        field_types = field_types or {}
        naming_context = naming_context or NamingContext()
        protocols: List[ProtocolSpec] = []
        warnings: List[str] = []

        existing_names = set(naming_context.existing_names)
        context = NamingContext(
            existing_names=existing_names,
            frequency=naming_context.frequency,
            fallback_prefix=naming_context.fallback_prefix,
        )

        for bundle, tier in bundle_tiers.items():
            bundle_set = set(bundle)
            if not self._bundle_allowed(bundle_set, tier):
                continue
            name = suggest_name(bundle_set, context)
            context.existing_names.add(name)
            fields = self._build_fields(bundle_set, field_types)
            protocols.append(
                ProtocolSpec(
                    name=name,
                    fields=fields,
                    bundle=bundle_set,
                    tier=tier,
                    rationale=f"Tier-{tier} bundle",
                )
            )

        if not protocols:
            warnings.append("No bundles qualified for synthesis.")

        return SynthesisPlan(protocols=protocols, warnings=warnings, errors=[])

    def _bundle_allowed(self, bundle: Set[str], tier: int) -> bool:
        if not bundle:
            return False
        if len(bundle) < self.config.min_bundle_size and not self.config.allow_singletons:
            return False
        return tier <= self.config.max_tier

    def _build_fields(
        self, bundle: Iterable[str], field_types: Dict[str, str]
    ) -> List[FieldSpec]:
        fields: List[FieldSpec] = []
        for name in sorted(bundle):
            type_hint = field_types.get(name)
            fields.append(FieldSpec(name=name, type_hint=type_hint, source_params={name}))
        return fields



from __future__ import annotations

from pathlib import Path
import sys
import textwrap


def _write(tmp_path: Path, rel: str, content: str) -> Path:
    path = tmp_path / rel
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(textwrap.dedent(content).lstrip())
    return path


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import AuditConfig, build_synthesis_plan

    return AuditConfig, build_synthesis_plan


def test_synthesis_plan_applies_tiers_and_max_tier(tmp_path: Path) -> None:
    AuditConfig, build_synthesis_plan = _load()
    _write(
        tmp_path,
        "settings.py",
        """
        from dataclasses import dataclass

        @dataclass
        class AppConfig:
            a: int
            b: int
        """,
    )
    groups_by_path = {
        tmp_path / "mod.py": {
            "f": [{"a", "b"}],
            "g": [{"c", "d"}],
            "h": [{"c", "d"}],
            "i": [{"e", "f"}],
        }
    }
    plan = build_synthesis_plan(
        groups_by_path,
        project_root=tmp_path,
        max_tier=2,
        config=AuditConfig(project_root=tmp_path),
    )
    protocols = plan.get("protocols", [])
    tiers = {tuple(sorted(p.get("bundle", []))): p.get("tier") for p in protocols}
    assert tiers.get(("a", "b")) == 1
    assert tiers.get(("c", "d")) == 2
    assert ("e", "f") not in tiers



from __future__ import annotations

from pathlib import Path

import libcst as cst

from gabion.refactor import engine as refactor_engine


def test_import_helpers_and_insert_index() -> None:
    module = cst.parse_module('"""Doc"""\nimport typing\nfrom typing import Protocol\n')
    body = list(module.body)
    assert refactor_engine._is_docstring(body[0]) is True
    assert refactor_engine._is_import(body[1]) is True
    assert refactor_engine._find_import_insert_index(body) == 3
    assert refactor_engine._has_typing_import(body) is True
    assert refactor_engine._has_typing_protocol_import(body) is True
    assert refactor_engine._has_typing_overload_import(body) is False
    assert refactor_engine._has_warnings_import(body) is False

    module = refactor_engine._ensure_compat_imports(module)
    new_body = list(module.body)
    assert refactor_engine._has_typing_overload_import(new_body) is True
    assert refactor_engine._has_warnings_import(new_body) is True


def test_module_expr_to_str() -> None:
    assert refactor_engine._module_expr_to_str(cst.Name("typing")) == "typing"
    expr = cst.Attribute(cst.Name("a"), cst.Name("b"))
    assert refactor_engine._module_expr_to_str(expr) == "a.b"
    assert refactor_engine._module_expr_to_str(None) is None


def test_collect_import_context() -> None:
    module = cst.parse_module(
        "import pkg.mod as pm\n"
        "from pkg.mod import Target as Alias, Protocol as Proto\n"
        "from other import Thing\n"
    )
    module_aliases, imported_targets, protocol_alias = refactor_engine._collect_import_context(
        module, target_module="pkg.mod", protocol_name="Protocol"
    )
    assert module_aliases == {"pm": "pkg.mod"}
    assert imported_targets == {"Alias": "Target", "Proto": "Protocol"}
    assert protocol_alias == "Proto"


def test_rewrite_call_sites_target_and_imports(tmp_path: Path) -> None:
    source = (
        "from pkg.mod import target\n"
        "def f(a, b):\n"
        "    target(a, b)\n"
        "    target(a)\n"
        "    target(a, b, c)\n"
        "    target(a, c=1)\n"
        "    target(a, b=b)\n"
        "    target(Bundle(a=a, b=b))\n"
        "    target(*args)\n"
    )
    module = cst.parse_module(source)
    warnings, updated = refactor_engine._rewrite_call_sites(
        module,
        file_path=tmp_path / "consumer.py",
        target_path=tmp_path / "consumer.py",
        target_module="pkg.mod",
        protocol_name="Bundle",
        bundle_fields=["a", "b"],
        targets={"target"},
    )
    assert updated is not None
    assert any("extra positional" in w for w in warnings)
    assert any("unknown keyword" in w for w in warnings)
    assert any("missing bundle fields" in w for w in warnings)
    assert any("star args" in w for w in warnings)

    source = "from pkg.mod import target\n\ndef f(a, b):\n    target(a, b)\n"
    module = cst.parse_module(source)
    warnings, updated = refactor_engine._rewrite_call_sites(
        module,
        file_path=tmp_path / "other.py",
        target_path=tmp_path / "target.py",
        target_module="pkg.mod",
        protocol_name="Bundle",
        bundle_fields=["a", "b"],
        targets={"target"},
    )
    assert updated is not None
    assert "from pkg.mod import Bundle" in updated.code


def test_rewrite_call_sites_in_project(tmp_path: Path) -> None:
    target = tmp_path / "src" / "target.py"
    target.parent.mkdir(parents=True, exist_ok=True)
    target.write_text("def target(a, b):\n    return a + b\n")
    bad = tmp_path / "src" / "bad.py"
    bad.write_text("def oops(\n")
    consumer = tmp_path / "src" / "consumer.py"
    consumer.write_text("from target import target\n\ndef f(a, b):\n    target(a, b)\n")
    edits, warnings = refactor_engine._rewrite_call_sites_in_project(
        project_root=tmp_path,
        target_path=target,
        target_module="target",
        protocol_name="Bundle",
        bundle_fields=["a", "b"],
        targets={"target"},
    )
    assert warnings
    assert edits



from __future__ import annotations

from pathlib import Path
import sys


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import AuditConfig, analyze_paths

    return AuditConfig, analyze_paths


def test_type_flow_suggestions_and_ambiguities(tmp_path: Path) -> None:
    AuditConfig, analyze_paths = _load()
    code = (
        "def callee_int(x: int):\n"
        "    return x\n"
        "\n"
        "def callee_str(x: str):\n"
        "    return x\n"
        "\n"
        "def caller(a):\n"
        "    return callee_int(a)\n"
        "\n"
        "def caller_conflict(b):\n"
        "    callee_int(b)\n"
        "    callee_str(b)\n"
    )
    path = tmp_path / "mod.py"
    path.write_text(code)
    config = AuditConfig(
        project_root=tmp_path,
        exclude_dirs=set(),
        ignore_params=set(),
        external_filter=False,
        strictness="high",
        transparent_decorators=None,
    )
    analysis = analyze_paths(
        [path],
        recursive=False,
        type_audit=True,
        type_audit_report=True,
        type_audit_max=10,
        include_constant_smells=False,
        include_unused_arg_smells=False,
        config=config,
    )
    assert any("caller.a" in entry for entry in analysis.type_suggestions)
    assert any("caller_conflict.b" in entry for entry in analysis.type_ambiguities)



from __future__ import annotations

from pathlib import Path

import pytest
import typer

from gabion import cli


def test_split_csv_helpers() -> None:
    assert cli._split_csv_entries(None) is None
    assert cli._split_csv_entries(["a, b", " ", "c"]) == ["a", "b", "c"]
    assert cli._split_csv_entries([" ", ""]) is None

    assert cli._split_csv(None) is None
    assert cli._split_csv("a, , b") == ["a", "b"]
    assert cli._split_csv(" ,") is None


def test_build_refactor_payload_input_payload_passthrough() -> None:
    payload = {"protocol_name": "Bundle", "bundle": ["a"]}
    assert cli.build_refactor_payload(
        input_payload=payload,
        protocol_name=None,
        bundle=None,
        field=None,
        target_path=None,
        target_functions=None,
        compatibility_shim=False,
        rationale=None,
    ) == payload


def test_build_refactor_payload_requires_fields(tmp_path: Path) -> None:
    with pytest.raises(typer.BadParameter):
        cli.build_refactor_payload(
            protocol_name=None,
            bundle=None,
            field=None,
            target_path=None,
            target_functions=None,
            compatibility_shim=False,
            rationale=None,
        )

    payload = cli.build_refactor_payload(
        protocol_name="Bundle",
        bundle=None,
        field=[" : ", "a:int", "b"],
        target_path=tmp_path / "target.py",
        target_functions=None,
        compatibility_shim=False,
        rationale=None,
    )
    assert payload["bundle"] == ["a", "b"]
    assert payload["fields"] == [
        {"name": "a", "type_hint": "int"},
        {"name": "b", "type_hint": None},
    ]


def test_run_docflow_audit_missing_script(tmp_path: Path) -> None:
    missing = tmp_path / "missing.py"
    exit_code = cli._run_docflow_audit(
        root=tmp_path,
        fail_on_violations=False,
        script=missing,
    )
    assert exit_code == 2


def test_run_docflow_audit_passes_flags(tmp_path: Path) -> None:
    script = tmp_path / "docflow.py"
    script.write_text("import sys\nsys.exit(0)\n")
    exit_code = cli._run_docflow_audit(
        root=tmp_path,
        fail_on_violations=True,
        script=script,
    )
    assert exit_code == 0


def test_dataflow_audit_skips_type_audit_output() -> None:
    class DummyCtx:
        args: list[str] = []

    def runner(*_args, **_kwargs):
        # dataflow-bundle: _args, _kwargs
        return {"exit_code": 0}

    request = cli.DataflowAuditRequest(ctx=DummyCtx(), args=["sample.py"], runner=runner)
    with pytest.raises(typer.Exit) as exc:
        cli._dataflow_audit(request)
    assert exc.value.exit_code == 0


def test_dataflow_audit_type_audit_empty_findings() -> None:
    class DummyCtx:
        args: list[str] = []

    def runner(*_args, **_kwargs):
        # dataflow-bundle: _args, _kwargs
        return {"exit_code": 0, "type_suggestions": [], "type_ambiguities": []}

    request = cli.DataflowAuditRequest(
        ctx=DummyCtx(),
        args=["sample.py", "--type-audit", "--type-audit-max", "1"],
        runner=runner,
    )
    with pytest.raises(typer.Exit) as exc:
        cli._dataflow_audit(request)
    assert exc.value.exit_code == 0


def test_run_synth_parses_optional_inputs(tmp_path: Path) -> None:
    def runner(*_args, **_kwargs):
        # dataflow-bundle: _args, _kwargs
        return {"exit_code": 0}

    result, paths_out, timestamp = cli._run_synth(
        paths=None,
        root=tmp_path,
        out_dir=tmp_path / "out",
        no_timestamp=True,
        config=None,
        exclude=["a, b"],
        ignore_params="x, y",
        transparent_decorators="deco",
        allow_external=None,
        strictness=None,
        no_recursive=False,
        max_components=3,
        type_audit_report=True,
        type_audit_max=5,
        synthesis_max_tier=2,
        synthesis_min_bundle_size=1,
        synthesis_allow_singletons=False,
        synthesis_protocols_kind="dataclass",
        refactor_plan=False,
        fail_on_violations=False,
        runner=runner,
    )
    assert result["exit_code"] == 0
    assert timestamp is None
    assert paths_out["output_root"].exists()


def test_run_synthesis_plan_without_input(tmp_path: Path) -> None:
    captured = {}

    def runner(request, *, root=None):
        captured["request"] = request
        captured["root"] = root
        return {"protocols": []}

    output_path = tmp_path / "plan.json"
    cli._run_synthesis_plan(
        input_path=None,
        output_path=output_path,
        runner=runner,
    )
    assert captured["root"] is None
    assert output_path.read_text().strip()



from __future__ import annotations

from pathlib import Path
import sys


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.synthesis.model import NamingContext
    from gabion.synthesis.naming import _normalize_identifier, suggest_name

    return NamingContext, _normalize_identifier, suggest_name


def test_suggest_name_uses_frequency() -> None:
    NamingContext, _normalize_identifier, suggest_name = _load()
    context = NamingContext(frequency={"ctx": 3, "config": 1})
    name = suggest_name(["config", "ctx"], context)
    assert name == "CtxBundle"
    assert _normalize_identifier("$$$", "Fallback") == "Fallback"


def test_suggest_name_avoids_collisions() -> None:
    NamingContext, _normalize_identifier, suggest_name = _load()
    context = NamingContext(existing_names={"CtxBundle"}, frequency={"ctx": 2})
    name = suggest_name(["ctx"], context)
    assert name == "CtxBundle2"
    assert _normalize_identifier("1value", "X") == "X1value"


def test_suggest_name_with_empty_fields() -> None:
    NamingContext, _normalize_identifier, suggest_name = _load()
    context = NamingContext(fallback_prefix="Fallback")
    name = suggest_name([], context)
    assert name == "FallbackBundle"



from __future__ import annotations

from pathlib import Path
import sys


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion import server

    return server


def test_normalize_transparent_decorators() -> None:
    server = _load()
    assert server._normalize_transparent_decorators(None) is None
    assert server._normalize_transparent_decorators("a, b") == {"a", "b"}
    assert server._normalize_transparent_decorators(["a", "b, c"]) == {"a", "b", "c"}
    assert server._normalize_transparent_decorators([1, "a"]) == {"a"}
    assert server._normalize_transparent_decorators([]) is None


def test_uri_to_path() -> None:
    server = _load()
    path = Path("/tmp/demo.txt")
    assert server._uri_to_path(path.as_uri()) == path
    assert server._uri_to_path("relative/path.py") == Path("relative/path.py")


def test_diagnostics_for_path_reports_bundle(tmp_path: Path) -> None:
    server = _load()
    sample = tmp_path / "sample.py"
    sample.write_text(
        "def callee(x):\n"
        "    return x\n"
        "\n"
        "def caller(a, b):\n"
        "    callee(a)\n"
        "    callee(b)\n"
    )
    diagnostics = server._diagnostics_for_path(str(sample), tmp_path)
    assert diagnostics
    assert any("Implicit bundle" in diag.message for diag in diagnostics)


def test_start_uses_injected_callable() -> None:
    server = _load()
    called = {"value": False}

    def _start() -> None:
        called["value"] = True

    server.start(_start)
    assert called["value"] is True



"""Static analysis subpackage for Gabion."""

from .dataflow_audit import (
    AnalysisResult,
    AuditConfig,
    analyze_paths,
    apply_baseline,
    build_refactor_plan,
    build_synthesis_plan,
    compute_violations,
    load_baseline,
    render_dot,
    render_refactor_plan,
    render_report,
    render_protocol_stubs,
    render_synthesis_section,
    resolve_baseline_path,
    write_baseline,
)

__all__ = [
    "AnalysisResult",
    "AuditConfig",
    "analyze_paths",
    "apply_baseline",
    "build_refactor_plan",
    "build_synthesis_plan",
    "compute_violations",
    "load_baseline",
    "render_dot",
    "render_refactor_plan",
    "render_report",
    "render_protocol_stubs",
    "render_synthesis_section",
    "resolve_baseline_path",
    "write_baseline",
]



from __future__ import annotations

from pathlib import Path
import sys


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import (
        AuditConfig,
        analyze_paths,
        build_refactor_plan,
        build_synthesis_plan,
        render_dot,
        render_protocol_stubs,
        render_refactor_plan,
        render_report,
    )

    return (
        AuditConfig,
        analyze_paths,
        build_refactor_plan,
        build_synthesis_plan,
        render_dot,
        render_protocol_stubs,
        render_refactor_plan,
        render_report,
    )


def _write_kitchen_sink(root: Path) -> None:
    pkg = root / "pkg"
    pkg.mkdir(parents=True, exist_ok=True)
    (pkg / "__init__.py").write_text(
        "from .models import Config, UserConfig\n"
        "__all__ = [\"Config\", \"UserConfig\"]\n"
    )
    (pkg / "models.py").write_text(
        "from __future__ import annotations\n"
        "from dataclasses import dataclass\n"
        "from typing import Optional\n"
        "from .utils import helper\n"
        "\n"
        "@dataclass\n"
        "class Config:\n"
        "    name: str\n"
        "    count: int = 0\n"
        "\n"
        "@dataclass\n"
        "class UserConfig:\n"
        "    cfg: Config\n"
        "    tag: Optional[str] = None\n"
        "\n"
        "class Service:\n"
        "    def __init__(self, cfg: Config):\n"
        "        self.cfg = cfg\n"
        "\n"
        "    def run(self, a, b, *args, **kwargs):\n"
        "        return helper(a, b, *args, **kwargs)\n"
        "\n"
        "    @classmethod\n"
        "    def build(cls, cfg: Config):\n"
        "        return cls(cfg)\n"
        "\n"
        "    @staticmethod\n"
        "    def ping(x):\n"
        "        return x\n"
        "\n"
        "    @property\n"
        "    def value(self):\n"
        "        return self.cfg.count\n"
    )
    (pkg / "utils.py").write_text(
        "from __future__ import annotations\n"
        "\n"
        "def deco(fn):\n"
        "    def wrapper(*args, **kwargs):\n"
        "        return fn(*args, **kwargs)\n"
        "    return wrapper\n"
        "\n"
        "@deco\n"
        "def helper(a, b, *args, **kwargs):\n"
        "    # dataflow-bundle: a, b\n"
        "    c = a\n"
        "    d, e = (b, b)\n"
        "    data = {\"k\": c}\n"
        "    data[\"k\"] = d\n"
        "    return c, d\n"
        "\n"
        "def uses_kwargs(**kwargs):\n"
        "    return helper(**kwargs)\n"
    )
    (pkg / "calls.py").write_text(
        "from __future__ import annotations\n"
        "from .models import Config, Service\n"
        "\n"
        "def run(cfg: Config):\n"
        "    svc = Service.build(cfg)\n"
        "    return svc.run(cfg.name, cfg.count)\n"
    )
    (pkg / "nested.py").write_text(
        "def outer(x):\n"
        "    def inner(y):\n"
        "        return y\n"
        "    return inner(x)\n"
    )
    (pkg / "exports.py").write_text(
        "from .models import Config as Cfg, UserConfig\n"
        "__all__ = [\"Cfg\", \"UserConfig\"]\n"
    )
    (pkg / "star.py").write_text("from .exports import *\n")
    (root / "main.py").write_text(
        "from __future__ import annotations\n"
        "from dataclasses import dataclass\n"
        "from pkg.models import Config, UserConfig\n"
        "from pkg.calls import run\n"
        "\n"
        "@dataclass\n"
        "class LocalCfg:\n"
        "    value: int\n"
        "\n"
        "def entry():\n"
        "    cfg = Config(name=\"demo\", count=1)\n"
        "    user = UserConfig(cfg=cfg)\n"
        "    return run(cfg), user\n"
    )


def test_kitchen_sink_analysis_outputs(tmp_path: Path) -> None:
    (
        AuditConfig,
        analyze_paths,
        build_refactor_plan,
        build_synthesis_plan,
        render_dot,
        render_protocol_stubs,
        render_refactor_plan,
        render_report,
    ) = _load()
    _write_kitchen_sink(tmp_path)
    config = AuditConfig(
        project_root=tmp_path,
        exclude_dirs=set(),
        ignore_params=set(),
        external_filter=False,
        strictness="high",
        transparent_decorators={"deco"},
    )
    analysis = analyze_paths(
        [tmp_path],
        recursive=True,
        type_audit=True,
        type_audit_report=True,
        type_audit_max=5,
        include_constant_smells=True,
        include_unused_arg_smells=True,
        config=config,
    )
    report, violations = render_report(
        analysis.groups_by_path,
        5,
        type_suggestions=analysis.type_suggestions,
        type_ambiguities=analysis.type_ambiguities,
        constant_smells=analysis.constant_smells,
        unused_arg_smells=analysis.unused_arg_smells,
    )
    assert "Dataflow grammar" in report or "dataflow-grammar" in report
    assert isinstance(violations, list)

    dot = render_dot(analysis.groups_by_path)
    assert "digraph" in dot

    plan = build_synthesis_plan(
        analysis.groups_by_path,
        project_root=tmp_path,
        max_tier=2,
        min_bundle_size=1,
        allow_singletons=True,
        config=config,
    )
    stubs = render_protocol_stubs(plan, kind="dataclass")
    assert "class" in stubs

    refactor_plan = build_refactor_plan(
        analysis.groups_by_path,
        [tmp_path],
        config=config,
    )
    refactor_summary = render_refactor_plan(refactor_plan)
    assert "Refactoring plan" in refactor_summary
